#!/usr/bin/env python
# -*- coding: utf-8 -*-

import platform
import sys
from pathlib import Path
from typing import Dict, Union

import torch
from pythonwrench.json import dump_json

import torchwrench as tw
from torchwrench.core.packaging import _EXTRA_VERSION
from torchwrench.hub.paths import get_cache_dir, get_dir, get_tmp_dir
from torchwrench.utils.data.dataloader import get_auto_num_cpus, get_auto_num_gpus


def get_package_repository_path() -> str:
    """Return the absolute path where the source code of this package is installed."""
    return str(Path(__file__).parent.parent.parent)


def get_install_info() -> Dict[str, Union[str, int]]:
    install_info = {
        "torchwrench": tw.__version__,
        "python": f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
        "os": platform.system(),
        "architecture": platform.architecture()[0],
        "torch": str(torch.__version__),
        "num_cpus": get_auto_num_cpus(),
        "num_gpus": get_auto_num_gpus(),
        "package_path": get_package_repository_path(),
        "tmpdir": str(get_tmp_dir()),
        "cachedir": str(get_cache_dir()),
        "torch_hub": get_dir(),
    }
    install_info.update({k: str(v) for k, v in _EXTRA_VERSION.items()})
    return install_info


def print_install_info() -> None:
    """Show main packages versions."""
    install_info = get_install_info()
    dumped = dump_json(install_info)
    print(dumped)


if __name__ == "__main__":
    print_install_info()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os.path as osp
import pickle
import random
import string
import time
import unittest
import warnings
from unittest import TestCase

import numpy as np
import pythonwrench as pw
import torch
from pythonwrench import dict_list_to_list_dict
from torch import Tensor
from torchvision.datasets import CIFAR10

import torchwrench as tw
from torchwrench import nn
from torchwrench.extras.hdf import HDFDataset, pack_to_hdf
from torchwrench.extras.hdf.pack import hdf_dtype_to_numpy_dtype
from torchwrench.hub.paths import get_tmp_dir
from torchwrench.utils.data.dataset import Subset


class TestHDF(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        tmpdir = get_tmp_dir().joinpath("torchwrench_tests", "hdf")
        tmpdir.mkdir(parents=True, exist_ok=True)
        cls.tmpdir = tmpdir

    def test_cifar10_pack_to_hdf(self) -> None:
        cls = self.__class__
        tmpdir = cls.tmpdir

        dataset = CIFAR10(
            str(tmpdir),
            train=False,
            transform=nn.ToNDArray(),
            target_transform=nn.Sequential(nn.IndexToOnehot(10), nn.ToList()),
            download=True,
        )
        indices = torch.randint(0, len(dataset), (max(len(dataset) // 10, 1),)).tolist()
        dataset = Subset(dataset, indices)

        path = tmpdir.joinpath("test_cifar10_pack_to_hdf.hdf")
        hdf_dataset = pack_to_hdf(dataset, path, exists="overwrite")

        idx = 0
        image0, label0 = dataset[idx]
        image1, label1 = hdf_dataset[idx]

        assert len(dataset) == len(hdf_dataset)
        assert np.equal(image0, image1).all()
        assert np.equal(label0, label1).all()

        # Try pickle for DDP support
        dumped = pickle.dumps(hdf_dataset)
        hdf_dataset_2 = pickle.loads(dumped)
        assert hdf_dataset == hdf_dataset_2

        hdf_dataset_2.close(remove_file=False)
        hdf_dataset.close(remove_file=True)

    def test_shape_column(self) -> None:
        cls = self.__class__
        tmpdir = cls.tmpdir

        shape_suffix = "_added_shape"
        data = [torch.rand(10, 2), torch.rand(10, 5), torch.rand(10, 3)]
        data_shape = [(10, 1), (10, 2), (10, 1)]

        ds_dict = {"data": data, f"data{shape_suffix}": data_shape}
        ds_list = dict_list_to_list_dict(ds_dict, "same")

        path = tmpdir.joinpath("test_shape.hdf")
        hdf_dataset = pack_to_hdf(
            ds_list,
            path,
            exists="overwrite",
            shape_suffix=shape_suffix,
            ds_kwds=dict(cast="to_torch_or_builtin"),
        )

        assert len(hdf_dataset.added_columns) == 0
        assert len(hdf_dataset) == len(ds_list)

        for i, item in enumerate(iter(hdf_dataset)):
            assert set(item.keys()) == set(ds_dict.keys())

            data_i = item["data"]
            assert isinstance(data_i, Tensor)
            assert data_i.shape == data_shape[i], (
                f"{i=}; {data_i.shape=}; {data_shape[i]=}"
            )

        hdf_dataset.close(remove_file=True)

    def test_indices_mask(self) -> None:
        cls = self.__class__
        tmpdir = cls.tmpdir

        data = np.random.rand(100, 8)
        ds_dict = {"data": data}
        ds_list = dict_list_to_list_dict(ds_dict)

        path = tmpdir.joinpath("test_indices_slicing.hdf")
        hdf_dataset = pack_to_hdf(
            ds_list,
            path,
            exists="overwrite",
            ds_kwds=dict(cast="none"),
        )

        # Random indices test
        # with generate indices with duplicates in random order !
        indices = np.random.randint(0, len(data), (10,))
        assert (hdf_dataset[indices, "data"] == data[indices]).all()

        # Random mask test
        mask = np.random.rand(len(data)) > 0.2
        indices = np.where(mask)[0]

        expected = hdf_dataset[mask, "data"]
        for values in (
            hdf_dataset[indices, "data"],
            hdf_dataset[mask.tolist(), "data"],
            hdf_dataset[indices.tolist(), "data"],
            hdf_dataset[tw.as_tensor(mask), "data"],
            hdf_dataset[tw.as_tensor(indices), "data"],
            data[mask],
            data[indices],
        ):
            assert (expected == values).all()

        hdf_dataset.close(remove_file=True)

    def test_special_floating_dtypes(self) -> None:
        cls = self.__class__
        tmpdir = cls.tmpdir

        ds_dict = {
            "f16": torch.rand(10, 2, 3, dtype=torch.float16),
            "f32": torch.rand(10, 2, 3, dtype=torch.float32),
            "f64": torch.rand(10, 2, 3, dtype=torch.float64),
            "c64": torch.rand(10, 2, 3, dtype=torch.complex64),
            "c128": torch.rand(10, 2, 3, dtype=torch.complex128),
            "bool": torch.rand(10, 1) > 0.5,
        }
        ds_list = dict_list_to_list_dict(ds_dict, "same")
        keys = set(ds_dict.keys())

        fname = "test_special_floating_dtypes.hdf"
        fpath = tmpdir.joinpath(fname)
        hdf_ds = pack_to_hdf(
            ds_list,
            fpath,
            exists="overwrite",
            ds_kwds=dict(cast="to_torch_src"),
        )

        assert len(hdf_ds.added_columns) == 0
        assert len(hdf_ds) == len(ds_list)

        for i, hdf_item in enumerate(iter(hdf_ds)):
            assert set(hdf_item.keys()) == keys
            assert set(ds_list[i].keys()) == keys

            src_item = ds_list[i]
            for k in keys:
                msg = f"Index: {i}; Key: {k}; {src_item[k]=}; {hdf_item[k]=}"
                src_value = src_item[k]
                hdf_value = hdf_item[k]
                assert torch.equal(src_value, hdf_value), msg

        hdf_ds.close(remove_file=True)

    def test_slice(self) -> None:
        cls = self.__class__
        tmpdir = cls.tmpdir

        ds_dict = {
            "a": list(range(10, 20)),
            "b": torch.rand(10, 2, 3),
        }
        ds_list = dict_list_to_list_dict(ds_dict, "same")

        path = tmpdir.joinpath("test_slice.hdf")
        pack_to_hdf(ds_list, path, exists="overwrite", ds_kwds=dict(open_hdf=False))

        hdf_dataset = HDFDataset(path, cast="as_builtin")

        assert len(hdf_dataset) == len(ds_list)
        assert hdf_dataset[:, "a"] == ds_dict["a"]
        assert hdf_dataset[:, "b"] == ds_dict["b"].tolist()

        indices = torch.randperm(len(hdf_dataset))

        assert hdf_dataset[indices, "a"] == [ds_dict["a"][idx] for idx in indices]
        assert (hdf_dataset[indices, "b"] == ds_dict["b"][indices].numpy()).all()

    def test_string(self) -> None:
        cls = self.__class__
        tmpdir = cls.tmpdir

        ds_dict = {
            "int": list(range(10)),
            "string": [
                "".join(map(str, range(random.randint(10, 100)))) for _ in range(10)
            ],
            "list_string": [
                [],
                ["aa", "bbb"],
                [],
                ["cccc"],
                [""],
                ["dd", "", "e", "ff", "gggggg"],
                ["éééé", "û", "é"],
            ]
            + [[]] * 3,
            "empty_lists": [[]] * 10,
            "bytes": [b""] * 6 + [b"a2", b"dnqzu1dhqz", b"0djqizjdz", b"du12qzd"],
            # TODO: add support for bytearray ?
            "bytearray": np.array([bytearray(str(i), "utf-8") for i in range(10, 20)]),
        }
        ds_list = dict_list_to_list_dict(ds_dict, "same")

        path = tmpdir.joinpath("test_string.hdf")
        hdf_dataset = pack_to_hdf(
            ds_list,
            path,
            exists="overwrite",
            ds_kwds=dict(cast="as_builtin"),
        )

        assert hdf_dataset._src_is_unicode == {
            "int": False,
            "string": True,
            "list_string": True,
            "empty_lists": False,
            "bytes": False,
            "bytearray": False,
        }
        idx = torch.randint(0, len(hdf_dataset), ()).item()
        col = random.choice(list(ds_dict.keys()))
        eq = hdf_dataset[idx, col] == ds_dict[col][idx]
        if isinstance(eq, np.ndarray):
            eq = eq.all()
        assert eq, f"{idx=}; {col=}"

        assert len(hdf_dataset) == len(ds_list)
        for k in ds_dict.keys():
            hdf_col = hdf_dataset[:, k]
            ds_col = ds_dict[k]
            eq = hdf_col == ds_col
            if isinstance(eq, np.ndarray):
                eq = eq.all()
            assert eq, f"{k=}, {hdf_col=} != {ds_col=}"

    def test_string_comp(self) -> None:
        num_data = 10000
        max_string_len = 100
        max_sublst = 5
        seed = 42

        gen = torch.Generator().manual_seed(seed)
        # note: vlen_str does not support chr(0)
        strings = [
            "".join(
                map(
                    chr,
                    torch.randint(1, 256, (max_string_len,), generator=gen).tolist(),
                )
            )
            for _ in range(num_data)
        ]
        list_strings = [
            [
                "".join(
                    map(
                        chr,
                        torch.randint(
                            1, 256, (max_string_len,), generator=gen
                        ).tolist(),
                    )
                )
                for _ in range(1, max_sublst + 1)
            ]
            for _ in range(num_data)
        ]
        ds_dict = {"strings": strings, "list_strings": list_strings}
        ds_list = dict_list_to_list_dict(ds_dict, key_mode="same")

        ds_vlen = pack_to_hdf(
            ds_list,
            self.tmpdir.joinpath("test_string_comp_vlen.hdf"),
            exists="overwrite",
            store_str_as_vlen=True,
            verbose=2,
        )
        ds_bytes = pack_to_hdf(
            ds_list,
            self.tmpdir.joinpath("test_string_comp_bytes.hdf"),
            exists="overwrite",
            store_str_as_vlen=False,
            verbose=2,
        )

        n_try = 50
        duration_ds_vlen_lst = []
        duration_ds_bytes_lst = []

        for i in range(n_try):
            rand_val = random.randint(0, 1)
            ds_bytes_data = {}
            duration_ds_bytes = 0.0

            if i % 2 == rand_val:
                start = time.perf_counter()
                ds_bytes_data = ds_bytes[:]
                duration_ds_bytes = time.perf_counter() - start

            start = time.perf_counter()
            ds_vlen_data = ds_vlen[:]
            duration_ds_vlen = time.perf_counter() - start

            if i % 2 == (1 - rand_val):
                start = time.perf_counter()
                ds_bytes_data = ds_bytes[:]
                duration_ds_bytes = time.perf_counter() - start

            duration_ds_vlen_lst.append(duration_ds_vlen)
            duration_ds_bytes_lst.append(duration_ds_bytes)

            assert ds_bytes_data.keys() == ds_vlen_data.keys()
            assert all(
                np.all(ds_bytes_data[k] == ds_vlen_data[k])
                for k in ds_bytes_data.keys()
            )

        if not (np.median(duration_ds_bytes_lst) < np.median(duration_ds_vlen_lst)):
            warnings.warn("HDF read speed comparison failed")

        ds_vlen.close(remove_file=True)
        ds_bytes.close(remove_file=True)

    def test_conversion(self) -> None:
        assert hdf_dtype_to_numpy_dtype("i") == np.dtype("int32")

        with self.assertRaises(ValueError):
            assert hdf_dtype_to_numpy_dtype("invalid") == np.dtype("V")  # type: ignore

    def test_pack_dict(self) -> None:
        num_rows = tw.randint(100, 200, ()).item()
        data = {
            "x": tw.rand(num_rows, 10),
            "y": tw.randint(0, 1000, (num_rows, 10)),
            "z": [pw.randstr(10, letters=string.printable) for _ in range(num_rows)],
        }
        hdf_fpath = self.__class__.tmpdir.joinpath("test_pack_dict.hdf")
        ds = pack_to_hdf(data, hdf_fpath, exists="overwrite")

        assert tw.deep_equal(ds.to_dict(), data)

    def test_compression(self) -> None:
        num_rows = tw.randint(1000, 2000, ()).item()
        dimsize = 1000

        lengths = tw.randint(0, dimsize // 100, (num_rows,))
        mask = tw.lengths_to_non_pad_mask(lengths, dimsize)

        data = {
            "x": tw.where(mask, tw.rand(num_rows, dimsize), tw.as_tensor(0.0)),
        }

        hdf_fpath_uncompressed = self.__class__.tmpdir.joinpath("test_uncompressed.hdf")
        hdf_uncompressed = pack_to_hdf(data, hdf_fpath_uncompressed, exists="overwrite")

        hdf_fpath_compressed = self.__class__.tmpdir.joinpath("test_compressed.hdf")
        hdf_compressed = pack_to_hdf(
            data,
            hdf_fpath_compressed,
            col_kwds=dict(compression="lzf"),
            exists="overwrite",
        )

        assert tw.deep_equal(hdf_uncompressed[:], hdf_compressed[:])
        assert osp.getsize(hdf_fpath_uncompressed) > osp.getsize(hdf_fpath_compressed)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import pythonwrench as pw
import torch

import torchwrench as tw
from torchwrench.core.packaging import _NUMPY_AVAILABLE
from torchwrench.extras.numpy import (
    ndarray_to_tensor,
    np,
    numpy_is_complex,
    numpy_is_complex_dtype,
    numpy_item,
    numpy_topk,
    numpy_view_as_complex,
    numpy_view_as_real,
    tensor_to_ndarray,
)


class TestNumpy(TestCase):
    def test_example_1(self) -> None:
        if not _NUMPY_AVAILABLE:
            return None

        x_tensor = torch.rand(3, 4, 5)
        x_array = tensor_to_ndarray(x_tensor)
        result = ndarray_to_tensor(x_array)

        assert torch.equal(x_tensor, result)

    def test_complex(self) -> None:
        if not _NUMPY_AVAILABLE:
            return None

        complex_dtypes = [np.complex64, np.complex128]
        x_complex = [
            np.array(
                np.random.rand(1) * 1j,
                dtype=complex_dtypes[np.random.randint(0, len(complex_dtypes))],
            )
            for _ in range(1000)
        ]
        assert all(numpy_is_complex(xi) for xi in x_complex)
        assert all(numpy_is_complex_dtype(xi.dtype) for xi in x_complex)

        x_real = [numpy_view_as_real(xi) for xi in x_complex]
        assert all(not numpy_is_complex(xi) for xi in x_real)
        assert all(not numpy_is_complex_dtype(xi.dtype) for xi in x_real)

        result = [numpy_view_as_complex(xi) for xi in x_real]
        assert all(numpy_is_complex(xi) for xi in result)
        assert all(numpy_is_complex_dtype(xi.dtype) for xi in result)
        assert x_complex == result

    def test_numpy_item(self) -> None:
        x = np.array(1.2)
        result = numpy_item(x)
        assert isinstance(result, np.generic)

        x = np.array([1, 2])[1]
        result = numpy_item(x)
        assert isinstance(result, np.generic)

        x = 10.99
        result = numpy_item(x)
        assert isinstance(result, np.generic)

    def test_numpy_topk(self) -> None:
        x = np.array([0, 2, 0, 1, 2])
        result = numpy_topk(x, k=3)
        expected = np.array([2, 2, 1]), np.array([1, 4, 3])
        assert tw.deep_equal(result, expected)


class TestReduceCompat(TestCase):
    def test_example_1(self) -> None:
        x1 = np.random.rand(10) > 0.5
        x2 = np.random.rand(10) > 0.5
        expected = x1 | x2

        result = pw.reduce_or(x1, x2)
        assert tw.deep_equal(result, expected)

        result = pw.reduce_or([x1, x2], start=False)
        assert tw.deep_equal(result, expected)

        result = pw.reduce_or(np.array([x1, x2]), start=False)
        assert tw.deep_equal(result, expected)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
from io import BufferedWriter
from pathlib import Path
from typing import BinaryIO, Optional, Tuple, TypedDict, Union

from pythonwrench._core import _setup_output_fpath
from pythonwrench.functools import function_alias
from pythonwrench.importlib import Placeholder
from torch import Tensor

from torchwrench.core.packaging import _TORCHAUDIO_AVAILABLE

if not _TORCHAUDIO_AVAILABLE:
    msg = f"Cannot use python module {__file__} since torchaudio package is not installed."
    raise ImportError(msg)

import torchaudio

try:
    from torchaudio import _AudioMetaData
    from torchaudio.io import CodecConfig  # type: ignore
except (ImportError, AttributeError):

    class CodecConfig(Placeholder): ...

    class _AudioMetaData(Placeholder): ...


class AudioMetaDataDict(TypedDict):
    sample_rate: int
    num_frames: int
    num_channels: int
    bits_per_sample: int
    encoding: str


def audio_metadata_to_dict(meta: _AudioMetaData) -> AudioMetaDataDict:
    return {
        "sample_rate": meta.sample_rate,
        "num_frames": meta.num_frames,
        "num_channels": meta.num_channels,
        "bits_per_sample": meta.bits_per_sample,
        "encoding": meta.encoding,
    }


def dump_with_torchaudio(
    src: Tensor,
    uri: Union[BinaryIO, str, Path, os.PathLike, None],
    sample_rate: int,
    channels_first: bool = True,
    format: Optional[str] = None,
    encoding: Optional[str] = None,
    bits_per_sample: Optional[int] = None,
    buffer_size: int = 4096,
    backend: Optional[str] = None,
    compression: Optional[Union[CodecConfig, float, int]] = None,
    *,
    overwrite: bool = True,
    make_parents: bool = True,
) -> bytes:
    """Dump tensors to audio waveform file. Requires torchaudio package installed."""
    if sample_rate <= 0:
        msg = f"Invalid argument {sample_rate=}. (expected positive value)"
        raise ValueError(msg)

    if isinstance(uri, (str, Path, os.PathLike)) or uri is None:
        uri = _setup_output_fpath(uri, overwrite, make_parents)

    buffer = io.BytesIO()
    torchaudio.save(  # type: ignore
        buffer,
        src,
        sample_rate,
        channels_first,
        format,
        encoding,
        bits_per_sample,
        buffer_size,
        backend,
        compression,  # type: ignore
    )
    content = buffer.getvalue()

    if isinstance(uri, Path):
        uri.write_bytes(content)
    elif isinstance(uri, (BinaryIO, BufferedWriter)):
        uri.write(content)
        uri.flush()

    return content


def load_with_torchaudio(
    uri: Union[BinaryIO, str, os.PathLike, Path],
    frame_offset: int = 0,
    num_frames: int = -1,
    normalize: bool = True,
    channels_first: bool = True,
    format: Optional[str] = None,
    buffer_size: int = 4096,
    backend: Optional[str] = None,
) -> Tuple[Tensor, int]:
    return torchaudio.load(  # type: ignore
        uri,
        frame_offset,
        num_frames,
        normalize,
        channels_first,
        format,
        buffer_size,
        backend,
    )


@function_alias(dump_with_torchaudio)
def dump_audio(*args, **kwargs): ...


@function_alias(load_with_torchaudio)
def load_audio(*args, **kwargs): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import tempfile
from pathlib import Path

from pythonwrench.functools import function_alias
from torch.hub import get_dir


def get_tmp_dir(mkdir: bool = False, make_parents: bool = True) -> Path:
    """Returns torchwrench temporary directory.

    Defaults is `/tmp/torchwrench`.
    Can be overriden with 'TORCHWRENCH_TMPDIR' environment variable.
    """
    default = tempfile.gettempdir()
    result = os.getenv("TORCHWRENCH_TMPDIR", default)
    result = Path(result).joinpath("torchwrench").resolve().expanduser()
    if mkdir:
        result.mkdir(parents=make_parents, exist_ok=True)
    return result


def get_cache_dir(mkdir: bool = False, make_parents: bool = True) -> Path:
    """Returns torchwrench cache directory for storing checkpoints, data and models.

    Defaults is `~/.cache/torchwrench`.
    Can be overriden with 'TORCHWRENCH_CACHEDIR' environment variable.
    """
    default = Path.home().joinpath(".cache", "torchwrench")
    result = os.getenv("TORCHWRENCH_CACHEDIR", default)
    result = Path(result).resolve().expanduser()
    if mkdir:
        result.mkdir(parents=make_parents, exist_ok=True)
    return result


@function_alias(get_dir)
def get_torch_cache_dir(*args, **kwargs): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Type used for backward compatibility."""

from typing import Generic, NamedTuple

from pythonwrench.semver import Version
from torch import Tensor, __version__
from typing_extensions import TypeVar

T = TypeVar("T")
T_Values = TypeVar("T_Values", bound=Tensor, covariant=True)
T_Indices = TypeVar("T_Indices", bound=Tensor, covariant=True)


__all__ = [
    "max",
    "min",
    "sort",
    "topk",
    "shape",
    "ndim",
    "top_p",
]


class _namedtuple_values_indices(Generic[T_Values, T_Indices], tuple):
    @property
    def values(self) -> T_Values:
        return self[0]

    @property
    def indices(self) -> T_Indices:
        return self[1]


if Version(str(__version__)) >= Version("2.0.0"):
    from torch.return_types import max, min, sort, topk  # type: ignore # noqa: F401

else:

    class max(_namedtuple_values_indices[Tensor, Tensor]): ...

    class min(_namedtuple_values_indices[Tensor, Tensor]): ...

    class sort(_namedtuple_values_indices[Tensor, Tensor]): ...

    class topk(_namedtuple_values_indices[Tensor, Tensor]): ...


if Version.python() < (3, 11, 0):
    # workaround for typing in python 3.8 - 3.10
    class _indicator_base(NamedTuple):
        valid: bool
        shape: T  # type: ignore

    class shape(_indicator_base, Generic[T]):  # type: ignore
        ...

    class ndim(_indicator_base, Generic[T]):  # type: ignore
        ...

else:

    class shape(NamedTuple, Generic[T]):
        valid: bool
        shape: T

    class ndim(NamedTuple, Generic[T]):
        valid: bool
        ndim: T


class top_p(_namedtuple_values_indices): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from io import TextIOBase
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Tuple,
    Union,
)

import pythonwrench as pw
from pythonwrench.typing import SupportsGetitemIterLen
from torch import Tensor
from typing_extensions import Self

import torchwrench as tw
from torchwrench.extras.numpy import np
from torchwrench.extras.pandas import pd
from torchwrench.extras.speechbrain import DynamicItemDataset
from torchwrench.serialization.csv import read_csv, save_csv
from torchwrench.serialization.json import read_json, save_json

from ._core import (
    DataFrameWrapper,
    DictListWrapper,
    DynamicDatasetWrapper,
    FunctionWrapper,
    IndexerWrapper,
    T_ColIndex,
    T_RowIndex,
    TabularDatasetInterface,
    TensorOrArrayWrapper,
    _get_from_idx_indices_slice_mask,
)

__all__ = [
    "TabularDataset",
    "T_RowIndex",
    "T_ColIndex",
]


class TabularDataset(
    Generic[T_RowIndex, T_ColIndex],
    TabularDatasetInterface[T_RowIndex, T_ColIndex],
):
    def __init__(
        self,
        data: Union[
            Mapping[Any, pw.SupportsGetitemIterLen],
            pw.SupportsGetitemIterLen[Dict[Any, Any]],
            pd.DataFrame,
            Tensor,
            np.ndarray,
            DynamicItemDataset,
            TabularDatasetInterface[T_RowIndex, T_ColIndex],
        ],
        row_mapper: Union[Mapping[T_RowIndex, T_RowIndex], None] = None,
        col_mapper: Union[MutableMapping[T_ColIndex, T_ColIndex], None] = None,
        fns_list: Iterable[
            Tuple[
                Union[Tuple[T_ColIndex, ...], T_ColIndex],
                Union[Tuple[T_ColIndex, ...], T_ColIndex],
                Callable,
            ]
        ] = (),
    ) -> None:
        if isinstance(data, TabularDatasetInterface):
            wrapper = data
        elif pw.isinstance_generic(data, Mapping[Any, pw.SupportsGetitemIterLen]):
            wrapper = DictListWrapper(data)
        elif pw.isinstance_generic(data, pw.SupportsGetitemIterLen[Dict]):
            datadict = pw.list_dict_to_dict_list(data, "same")
            wrapper = DictListWrapper(datadict)
        elif isinstance(data, pd.DataFrame):
            wrapper = DataFrameWrapper(data)
        elif isinstance(data, (Tensor, np.ndarray)):
            wrapper = TensorOrArrayWrapper(data)
        elif isinstance(data, DynamicItemDataset):
            wrapper = DynamicDatasetWrapper(data)
        else:
            msg = f"Invalid argument type {type(data)}."
            raise TypeError(msg)

        fns_list = list(fns_list)
        if len(fns_list) > 0:
            wrapper = FunctionWrapper(
                wrapper,  # type: ignore
                fns_list,
            )

        super().__init__()
        self._wrapper = wrapper
        self._row_mapper = row_mapper
        self._col_mapper = col_mapper

    @property
    def _row_mapper_dict(self) -> Mapping[T_RowIndex, T_RowIndex]:
        if self._row_mapper is None:
            row_names = self._wrapper.row_names
            return dict(zip(row_names, row_names))  # type: ignore
        else:
            return self._row_mapper

    @property
    def _col_mapper_dict(self) -> MutableMapping[T_ColIndex, T_ColIndex]:
        if self._col_mapper is None:
            col_names = self._wrapper.column_names
            return dict(zip(col_names, col_names))
        else:
            return self._col_mapper

    @classmethod
    def from_csv(cls, fpath: Union[str, Path, TextIOBase], **kwds) -> Self:
        data = read_csv(fpath, **kwds)
        return cls(data)

    @classmethod
    def from_json(cls, fpath: Union[str, Path, TextIOBase], **kwds) -> Self:
        data = read_json(fpath, **kwds)
        return cls(data)

    def add_dynamic_column(
        self,
        fn: Callable,
        requires: Tuple[T_ColIndex, ...],
        provides: Union[T_ColIndex, Tuple[T_ColIndex, ...]],
        add_to_output_keys: bool = True,
    ) -> None:
        col_mapper_before = self._col_mapper_dict
        if isinstance(self._wrapper, FunctionWrapper):
            self._wrapper.add_dynamic_column(fn, requires, provides)
        else:
            self._wrapper = FunctionWrapper(
                self._wrapper,  # type: ignore
                [(requires, provides, fn)],
            )

        if add_to_output_keys and self._col_mapper is not None:
            if isinstance(provides, (str, int)):
                provides = [provides]  # type: ignore
            for provide in provides:  # type: ignore
                self._col_mapper[provide] = provide
        elif not add_to_output_keys and self._col_mapper is None:
            self._col_mapper = col_mapper_before

    @property
    def row_names(self) -> pw.SupportsGetitemIterLen:
        if self._row_mapper is None:
            return self._wrapper.row_names
        else:
            return tuple(self._row_mapper.keys())

    @property
    def column_names(self) -> SupportsGetitemIterLen:
        if self._col_mapper is None:
            return self._wrapper.column_names
        else:
            return tuple(self._col_mapper.keys())

    def to_dataframe(self) -> pd.DataFrame:
        list_dict = self.to_list_dict()
        return pd.DataFrame(list_dict)

    def to_dict_list(self) -> Dict[T_ColIndex, List]:
        list_dict = self.to_list_dict()
        return pw.list_dict_to_dict_list(list_dict, "same")

    def to_list_dict(self) -> List[Dict[T_ColIndex, Any]]:
        datalist = self._wrapper.to_list_dict()
        datalist = [
            {
                tgt_col: datalist[src_row][src_col]
                for tgt_col, src_col in self._col_mapper_dict.items()
            }
            for tgt_row, src_row in self._row_mapper_dict.items()
        ]
        return datalist

    def to_numpy(self) -> np.ndarray:
        datalist = self.to_list_dict()
        return np.array([list(item.values()) for item in datalist])

    def to_tensor(self) -> Tensor:
        datalist = self.to_list_dict()
        return tw.as_tensor([list(item.values()) for item in datalist])

    def to_csv(self, fpath: Union[str, Path], *args, **kwargs) -> None:
        data = self.to_dict_list()
        save_csv(data, fpath, *args, **kwargs)

    def to_json(self, fpath: Union[str, Path], *args, **kwargs) -> None:
        data = self.to_dict_list()
        save_json(data, fpath, *args, **kwargs)

    def __getitem__(self, indexer_, /) -> Any:
        indexer = IndexerWrapper(indexer_, self)
        del indexer_

        if self._row_mapper is None:
            row_indexer = indexer.row
        else:
            row_indexer = _get_from_idx_indices_slice_mask(
                self._row_mapper,  # type: ignore
                indexer.row,
            )

        if indexer.has_col_indexer:
            if self._col_mapper is None:
                col_indexer = indexer.col
            elif indexer.single_col:
                col_indexer = self._col_mapper[indexer.col]  # type: ignore
            else:
                col_indexer = [self._col_mapper[col] for col in indexer.col]  # type: ignore

            result = self._wrapper[row_indexer, col_indexer]  # type: ignore
        else:
            result = self._wrapper[row_indexer]  # type: ignore

        if self._col_mapper is None or indexer.has_col_indexer:
            pass
        elif indexer.single_row:
            result = {tgt: result[src] for tgt, src in self._col_mapper.items()}  # type: ignore
        else:
            result = [
                {tgt: result_i[src] for tgt, src in self._col_mapper.items()}  # type: ignore
                for result_i in result
            ]

        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import subprocess
from subprocess import CalledProcessError
from typing import TypeVar, Union

T = TypeVar("T")


def get_githash_short(*, default: T = "unknown") -> Union[str, T]:
    cmd = ["git", "rev-parse", "--short", "HEAD"]
    try:
        return subprocess.check_output(cmd).decode().strip()
    except (CalledProcessError, PermissionError, FileNotFoundError):
        return default


def get_githash_full(*, default: T = "unknown") -> Union[str, T]:
    cmd = ["git", "rev-parse", "HEAD"]
    try:
        return subprocess.check_output(cmd).decode().strip()
    except (CalledProcessError, PermissionError, FileNotFoundError):
        return default



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any

import torch
from pythonwrench.typing import (
    is_builtin_number,
    is_builtin_scalar,
)
from torch import Tensor
from typing_extensions import TypeGuard, TypeIs

from torchwrench.core.make import DTypeLike, as_dtype
from torchwrench.extras.numpy import is_numpy_number_like, is_numpy_scalar_like, np

from ._typing import (
    IntegralTensor,
    NumberLike,
    ScalarLike,
    Tensor0D,
    TensorOrArray,
)


def is_number_like(x: Any) -> TypeGuard[NumberLike]:
    """Returns True if input is a scalar number.

    Accepted numbers-like objects are:
    - Python numbers (int, float, bool, complex)
    - Numpy zero-dimensional arrays
    - Numpy numbers
    - PyTorch zero-dimensional tensors
    """
    return is_builtin_number(x) or is_numpy_number_like(x) or isinstance(x, Tensor0D)


def is_scalar_like(x: Any) -> TypeGuard[ScalarLike]:
    """Returns True if input is a scalar number.

    Accepted scalar-like objects are:
    - Python scalars like (int, float, bool, complex, None, str, bytes)
    - Numpy zero-dimensional arrays
    - Numpy generic
    - PyTorch zero-dimensional tensors
    """
    return is_builtin_scalar(x) or is_numpy_scalar_like(x) or isinstance(x, Tensor0D)


def is_tensor_or_array(x: Any) -> TypeIs[TensorOrArray]:
    return isinstance(x, (Tensor, np.ndarray))


def is_integral_dtype(dtype: DTypeLike) -> bool:
    dtype = as_dtype(dtype)
    return isinstance(torch.empty((0,), dtype=dtype), IntegralTensor)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Iterable, Union

from pythonwrench.collections import dump_dict
from torch import Tensor

from torchwrench.core.make import GeneratorLike
from torchwrench.nn.functional.cropping import CropAlign, crop_dim, crop_dims

from .module import Module


class CropDim(Module):
    """
    For more information, see :func:`~torchwrench.nn.functional.crop.crop_dim`.
    """

    def __init__(
        self,
        target_length: int,
        *,
        align: CropAlign = "left",
        dim: int = -1,
        generator: GeneratorLike = None,
    ) -> None:
        super().__init__()
        self.target_length = target_length
        self.align: CropAlign = align
        self.dim = dim
        self.generator: GeneratorLike = generator

    def forward(self, x: Tensor) -> Tensor:
        return crop_dim(
            x,
            self.target_length,
            dim=self.dim,
            align=self.align,
            generator=self.generator,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                target_length=self.target_length,
                dim=self.dim,
                align=self.align,
            )
        )


class CropDims(Module):
    """
    For more information, see :func:`~torchwrench.nn.functional.crop.crop_dims`.
    """

    def __init__(
        self,
        target_lengths: Iterable[int],
        *,
        aligns: Union[CropAlign, Iterable[CropAlign]] = "left",
        dims: Iterable[int] = (-1,),
        generator: GeneratorLike = None,
    ) -> None:
        super().__init__()
        self.target_lengths = target_lengths
        self.dims = dims
        self.aligns: Union[CropAlign, Iterable[CropAlign]] = aligns
        self.generator: GeneratorLike = generator

    def forward(self, x: Tensor) -> Tensor:
        return crop_dims(
            x,
            self.target_lengths,
            dims=self.dims,
            aligns=self.aligns,
            generator=self.generator,
        )

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                target_lengths=self.target_lengths,
                dims=self.dims,
                aligns=self.aligns,
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abc import abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterable,
    List,
    Mapping,
    Tuple,
    Union,
    overload,
)

from torchwrench.extras.numpy import np
import pythonwrench as pw
import torch
from pythonwrench.typing import SupportsGetitemIterLen
from torch import Tensor
from torch.utils.data.dataset import Dataset
from typing_extensions import TypeVar

from torchwrench.extras.numpy import is_numpy_bool_array, is_numpy_integral_array
from torchwrench.extras.pandas import pd
from torchwrench.extras.speechbrain import DynamicItemDataset
from torchwrench.nn.functional.multilabel import multihot_to_multi_indices
from torchwrench.types import (
    BoolTensor1D,
    IntegralTensor,
    SignedIntegerTensor0D,
    SignedIntegerTensor1D,
)

T_RowIndex = TypeVar("T_RowIndex", bound=int, covariant=False, default=int)
T_ColIndex = TypeVar("T_ColIndex", bound=Union[int, str], covariant=False, default=str)


class SizedGenerator:
    def __init__(self, generator: Generator, size: int) -> None:
        super().__init__()
        self._generator = generator
        self._size = size

    def __iter__(self):
        yield from self._generator

    def __len__(self) -> int:
        return self._size


class TabularDatasetInterface(Dataset, Generic[T_RowIndex, T_ColIndex]):
    def __init__(self) -> None:
        super().__init__()

    @property
    def num_rows(self) -> int:
        return len(self.row_names)

    @property
    def num_columns(self) -> int:
        return len(self.column_names)

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.num_rows, self.num_columns

    @property
    def ndim(self) -> int:
        return len(self.shape)

    def keys(self) -> pw.SupportsGetitemIterLen[T_ColIndex]:
        return self.column_names

    def values(self) -> SizedGenerator:
        return SizedGenerator((self[k] for k in self.keys()), self.num_columns)

    def concat_columns_with(
        self,
        other: "TabularDatasetInterface[T_RowIndex, T_ColIndex]",
    ) -> "ColumnConcatWrapper[T_RowIndex, T_ColIndex]":
        return ColumnConcatWrapper([self, other])

    def __iter__(self):
        for row_index in self.row_names:
            yield self[row_index]

    def __len__(self) -> int:
        return self.num_rows

    @property
    @abstractmethod
    def row_names(self) -> SupportsGetitemIterLen[T_RowIndex]:
        raise NotImplementedError

    @property
    @abstractmethod
    def column_names(self) -> SupportsGetitemIterLen[T_ColIndex]:
        raise NotImplementedError

    @abstractmethod
    def to_dataframe(self) -> pd.DataFrame:
        raise NotImplementedError

    @abstractmethod
    def to_dict_list(self) -> Dict[T_ColIndex, List]:
        raise NotImplementedError

    @abstractmethod
    def to_list_dict(self) -> List[Dict[T_ColIndex, Any]]:
        raise NotImplementedError

    @abstractmethod
    def to_numpy(self) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def to_tensor(self) -> Tensor:
        raise NotImplementedError

    @overload
    def __getitem__(self, indexer: T_RowIndex, /) -> Dict[T_ColIndex, Any]: ...

    @overload
    def __getitem__(
        self,
        indexer: Union[
            Iterable[T_RowIndex], Iterable[T_ColIndex], Iterable[bool], slice
        ],
        /,
    ) -> List[Dict[T_ColIndex, Any]]: ...

    @overload
    def __getitem__(
        self, indexer: T_ColIndex, /
    ) -> Union[List[Any], np.ndarray, Tensor]: ...

    @overload
    def __getitem__(self, indexer: Tuple[T_RowIndex, T_ColIndex], /) -> Any: ...

    @abstractmethod
    def __getitem__(self, indexer) -> Any:
        raise NotImplementedError


class DictListWrapper(Generic[T_ColIndex], TabularDatasetInterface[int, T_ColIndex]):
    def __init__(self, data: Mapping[T_ColIndex, pw.SupportsGetitemIterLen]) -> None:
        lengths = list(map(len, data.values()))
        if not pw.all_eq(lengths):
            msg = f"Invalid data dict lengths. (found {lengths})"
            raise ValueError(msg)

        super().__init__()
        self._data = data

    @property
    def row_names(self) -> range:
        if len(self._data) == 0:
            return range(0)
        else:
            return range(len(next(iter(self._data.values()))))

    @property
    def column_names(self) -> tuple:
        return tuple(self._data.keys())

    def to_dataframe(self) -> pd.DataFrame:
        return pd.DataFrame(self._data)  # type: ignore

    def to_dict_list(self) -> Dict[T_ColIndex, List]:
        return {k: list(v) for k, v in self._data.items()}

    def to_list_dict(self) -> List[Dict[T_ColIndex, Any]]:
        return pw.dict_list_to_list_dict(self._data, "same")

    def to_numpy(self) -> np.ndarray:
        return np.array(list(self._data.values())).T

    def to_tensor(self) -> Tensor:
        return torch.as_tensor(list(self._data.values())).T

    def __getitem__(self, indexer_, /):  # type: ignore
        indexer = IndexerWrapper(indexer_, self)
        del indexer_

        if indexer.single_col:
            result = self._data[indexer.col]  # type: ignore
            result = _get_from_idx_indices_slice_mask(result, indexer.row)
            return result

        result_dict = {
            col: _get_from_idx_indices_slice_mask(self._data[col], indexer.row)
            for col in indexer.col  # type: ignore
        }

        if indexer.has_col_indexer:
            return list(result_dict.values())
        elif indexer.single_row:
            return result_dict
        else:
            result_list = pw.dict_list_to_list_dict(result_dict, "same")
            return result_list


class DataFrameWrapper(TabularDatasetInterface[int, str]):
    def __init__(self, data: pd.DataFrame) -> None:
        super().__init__()
        self._data = data

    @property
    def row_names(self) -> range:
        return range(self._data.shape[0])

    @property
    def column_names(self) -> tuple:
        return tuple(self._data.keys())

    def to_dataframe(self) -> pd.DataFrame:
        return self._data

    def to_dict_list(self) -> dict:
        return self._data.to_dict("list")

    def to_list_dict(self) -> list:
        return self._data.to_dict("records")

    def to_numpy(self) -> np.ndarray:
        return self._data.to_numpy()

    def to_tensor(self) -> Tensor:
        return torch.from_numpy(self._data.to_numpy())

    def __getitem__(self, indexer, /):
        return self._data[indexer]


class TensorOrArrayWrapper(TabularDatasetInterface[int, int]):
    def __init__(self, data: Union[Tensor, np.ndarray]) -> None:
        if data.ndim < 2:
            msg = f"Invalid number of dimensions for TabularDataset. (expected at least 2 dims but found {data.ndim})"
            raise ValueError(msg)

        super().__init__()
        self._data = data

    @property
    def row_names(self) -> range:
        return range(self._data.shape[0])

    @property
    def column_names(self) -> range:
        return range(self._data.shape[1])

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._data.shape

    def to_dataframe(self) -> pd.DataFrame:
        data = self._data
        if isinstance(data, Tensor):
            data = data.numpy(force=True)
        return pd.DataFrame(data)

    def to_dict_list(self) -> Dict[Any, List]:
        return dict(zip(self.column_names, map(list, self._data.transpose(0, 1))))

    def to_dynamic_item_dataset(self, id_column_name: int = 0) -> DynamicItemDataset:
        del id_column_name
        data_ids = list(range(len(self._data)))
        data = {id_: self._data[i] for i, id_ in enumerate(data_ids)}
        dset = DynamicItemDataset(data)
        return dset

    def to_list_dict(self) -> List[dict]:
        return [dict(zip(self.column_names, data_i)) for data_i in self._data]

    def to_numpy(self) -> np.ndarray:
        data = self._data
        if isinstance(data, Tensor):
            data = data.numpy(force=True)
        return data

    def to_tensor(self) -> Tensor:
        data = self._data
        if isinstance(data, np.ndarray):
            data = torch.from_numpy(data)
        return data

    def __getitem__(self, indexer, /):  # type: ignore
        return self._data[indexer]


class DynamicDatasetWrapper(TabularDatasetInterface[int, str]):
    def __init__(self, data: DynamicItemDataset) -> None:
        super().__init__()
        self._data = data

    @property
    def row_names(self) -> range:
        return range(len(self._data))

    @property
    def column_names(self) -> Tuple[str, ...]:
        return tuple(self._data.pipeline.output_mapping.keys())

    def to_dataframe(self) -> pd.DataFrame:
        return pd.DataFrame(self[:])

    def to_dict_list(self) -> Dict[Any, List]:
        datalist = self[:]
        return pw.list_dict_to_dict_list(datalist)  # type: ignore

    def to_list_dict(self) -> List[Dict]:
        datalist = self[:]
        return datalist  # type: ignore

    def to_numpy(self) -> np.ndarray:
        datalist = self[:]
        return np.array([list(item.values()) for item in datalist])  # type: ignore

    def to_tensor(self) -> Tensor:
        datalist = self[:]
        return torch.as_tensor([list(item.values()) for item in datalist])  # type: ignore

    def __getitem__(self, indexer_, /):  # type: ignore
        indexer = IndexerWrapper(indexer_, self)
        del indexer_

        output_keys = [indexer.col] if indexer.single_col else indexer.col
        with self._data.output_keys_as(output_keys):
            result = _get_from_idx_indices_slice_mask(self._data, indexer.row)

        if indexer.single_col and indexer.single_row:
            result = result[indexer.col]
            return result
        elif indexer.single_col and not indexer.single_row:
            result = [result_i[indexer.col] for result_i in result]
            return result
        elif indexer.has_col_indexer and indexer.single_row:
            result = list(result.values())
            return result
        elif indexer.has_col_indexer and not indexer.single_row:
            result = [list(result_i.values()) for result_i in result]
            return result
        elif not indexer.single_col:
            return result
        else:
            msg = f"Invalid state {indexer.single_row=}, {indexer.single_col=}, {indexer.has_col_indexer=}."
            raise TypeError(msg)


class FunctionWrapper(
    Generic[T_RowIndex, T_ColIndex],
    TabularDatasetInterface[T_RowIndex, T_ColIndex],
):
    def __init__(
        self,
        ds: TabularDatasetInterface[T_RowIndex, T_ColIndex],
        fns_list: Iterable[
            Tuple[
                Union[Tuple[T_ColIndex, ...], T_ColIndex],
                Union[Tuple[T_ColIndex, ...], T_ColIndex],
                Callable,
            ]
        ],
    ) -> None:
        fns = _get_fns_dict(fns_list)

        super().__init__()
        self._ds = ds
        self._fns = fns

    def add_dynamic_column(
        self,
        fn: Callable,
        requires: Tuple[T_ColIndex, ...],
        provides: Union[T_ColIndex, Tuple[T_ColIndex, ...]],
    ) -> None:
        provides_list = [provides] if isinstance(provides, (int, str)) else provides
        invalid = [provide for provide in provides_list if provide in self._fns]
        if len(invalid) > 0:
            msg = f"Found values already provided. (with {invalid=})"
            raise ValueError(msg)

        for provide in provides_list:
            self._fns[provide] = (requires, provides, fn)

    @property
    def row_names(self) -> range:
        return range(len(self._ds))

    @property
    def column_names(self) -> Tuple[T_ColIndex, ...]:
        return tuple(self._ds.column_names) + tuple(self._fns.keys())

    def to_dataframe(self) -> pd.DataFrame:
        return pd.DataFrame(self.to_dict_list())

    def to_dict_list(self) -> Dict[T_ColIndex, List]:
        datalist = self.to_list_dict()
        return pw.list_dict_to_dict_list(datalist)

    def to_list_dict(self) -> List[Dict[T_ColIndex, Any]]:
        datalist = self[:]
        return datalist  # type: ignore

    def to_numpy(self) -> np.ndarray:
        datalist = self.to_list_dict()
        return np.array([list(item.values()) for item in datalist])

    def to_tensor(self) -> Tensor:
        datalist = self.to_list_dict()
        return torch.as_tensor([list(item.values()) for item in datalist])

    def __getitem__(self, indexer_, /) -> Any:
        indexer = IndexerWrapper(indexer_, self)
        del indexer_

        row_indexer = indexer.row
        col_indexer: Iterable[T_ColIndex] = (
            [indexer.col] if indexer.single_col else indexer.col
        )  # type: ignore

        if isinstance(row_indexer, int):
            result = self._get_values(row_indexer, col_indexer)  # type: ignore
        elif isinstance(row_indexer, slice):
            row_indexer = range(len(self))[row_indexer]
            result = [self._get_values(idx, col_indexer) for idx in row_indexer]  # type: ignore
        elif pw.isinstance_generic(
            row_indexer, (Iterable[bool], BoolTensor1D)
        ) or is_numpy_bool_array(row_indexer):
            row_indexer = torch.as_tensor(row_indexer)
            row_indexer = multihot_to_multi_indices(row_indexer)
            result = [self._get_values(idx, col_indexer) for idx in row_indexer]  # type: ignore
        elif pw.isinstance_generic(
            row_indexer, (Iterable[int], SignedIntegerTensor1D)
        ) or is_numpy_integral_array(row_indexer):
            result = [self._get_values(idx, col_indexer) for idx in row_indexer]  # type: ignore
        else:
            msg = f"Invalid argument type {type(row_indexer)=}."
            raise TypeError(msg)

        if not indexer.has_col_indexer:
            return result
        elif indexer.single_row and indexer.single_col:
            return result[indexer.col]  # type: ignore
        elif indexer.single_row:
            return [result[col] for col in indexer.col]  # type: ignore
        elif indexer.single_col:
            return [result_i[indexer.col] for result_i in result]  # type: ignore
        else:
            return [[result_i[col] for col in indexer.col] for result_i in result]  # type: ignore

    def _get_values(
        self,
        idx: T_RowIndex,
        columns: Iterable[T_ColIndex],
    ) -> Dict[T_ColIndex, Any]:
        return _recursive_get_values(idx, columns, self._ds, self._fns)


def _get_fns_dict(
    fns_list: Iterable[
        Tuple[
            Union[Tuple[T_ColIndex, ...], T_ColIndex],
            Union[Tuple[T_ColIndex, ...], T_ColIndex],
            Callable,
        ]
    ],
) -> Dict[
    T_ColIndex,
    Tuple[
        Union[Tuple[T_ColIndex, ...], T_ColIndex],
        Union[Tuple[T_ColIndex, ...], T_ColIndex],
        Callable,
    ],
]:
    fns = {}
    for requires, provides, fn in fns_list:
        if isinstance(provides, (int, str)):
            fns[provides] = (requires, provides, fn)  # type: ignore
            continue

        for provide in provides:
            fns[provide] = (requires, provides, fn)

    return fns


def _recursive_get_values(
    idx: T_RowIndex,
    columns: Iterable[T_ColIndex],
    ds: TabularDatasetInterface[T_RowIndex, T_ColIndex],
    fns: Dict[T_ColIndex, Tuple],
) -> Dict[T_ColIndex, Any]:
    result = {}
    for col in columns:
        if col in result:
            continue

        if col in ds.column_names:
            result[col] = ds[idx, col]
            continue

        if col not in fns.keys():
            msg = f"Invalid argument {col=}. (expected one of {tuple(fns.keys())})"
            raise KeyError(msg)

        requires, provides, fn = fns[col]
        if isinstance(requires, (int, str)):
            requires = [requires]

        required_values_dict = _recursive_get_values(idx, requires, ds, fns)  # type: ignore
        required_values_list = [required_values_dict[k] for k in requires]
        provided_values_list = fn(*required_values_list)

        if isinstance(provides, (int, str)):
            provided_values_dict = {provides: provided_values_list}
        else:
            provided_values_dict = dict(zip(provides, provided_values_list))

        result.update(required_values_dict)
        result.update(provided_values_dict)

    return result


class ColumnConcatWrapper(
    Generic[T_RowIndex, T_ColIndex],
    TabularDatasetInterface[T_RowIndex, T_ColIndex],
):
    def __init__(
        self,
        tabulars: Iterable[TabularDatasetInterface[T_RowIndex, T_ColIndex]],
    ) -> None:
        tabulars = list(tabulars)

        super().__init__()
        self._dss = tabulars

    @property
    def row_names(self) -> SupportsGetitemIterLen:
        if len(self._dss) == 0:
            return []
        return self._dss[0].row_names

    @property
    def column_names(self) -> SupportsGetitemIterLen:
        return [col for tabular in self._dss for col in tabular.column_names]

    def to_dataframe(self) -> pd.DataFrame:
        return pd.concat(
            [tabular.to_dataframe() for tabular in self._dss], axis="columns"
        )

    def to_dict_list(self) -> Dict[Any, List]:
        return pw.reduce_or([tabular.to_dict_list() for tabular in self._dss], start={})

    def to_list_dict(self) -> List[Dict]:
        return pw.dict_list_to_list_dict(self.to_dict_list())

    def to_numpy(self) -> np.ndarray:
        return np.concat([tabular.to_numpy() for tabular in self._dss], axis=0)

    def to_tensor(self) -> Tensor:
        return torch.cat([tabular.to_tensor() for tabular in self._dss], dim=0)

    def __getitem__(self, indexer_, /) -> Any:
        indexer = IndexerWrapper(indexer_, self)
        del indexer_

        if indexer.single_col:
            for ds in self._dss:
                if indexer.col in ds.column_names:  # type: ignore
                    return ds[indexer.row, indexer.col]  # type: ignore
            raise TypeError

        result = []
        for ds in self._dss:
            ds_cols = [col for col in indexer.col if col in ds.column_names]  # type: ignore
            if len(ds_cols) == 0:
                continue

            result_i = ds[indexer.row, ds_cols]  # type: ignore
            result.append(result_i)

        if indexer.single_row:
            result = pw.reduce_or(result, start={})
            if indexer.has_col_indexer:
                return list(result.values())
            else:
                return result

        else:
            result_size = len(result[0]) if len(result) > 0 else 0
            result = [
                pw.reduce_or((result_i[j] for result_i in result), start={})
                for j in range(result_size)
            ]
            if indexer.has_col_indexer:
                return [list(result_i.values()) for result_i in result]
            else:
                return result


class IndexerWrapper:
    def __init__(
        self,
        indexer: Any,
        dataset: TabularDatasetInterface[Any, Any],
    ) -> None:
        if isinstance(indexer, IndexerWrapper):
            row_indexer = indexer._row_indexer
            col_indexer = indexer._col_indexer
            has_col_indexer = indexer._has_col_indexer

        else:
            if pw.isinstance_generic(
                indexer,
                (int, slice, Iterable[int], Iterable[bool], IntegralTensor, np.ndarray),
            ):
                row_indexer = indexer
                col_indexer = None
            elif isinstance(indexer, tuple) and len(indexer) == 2:
                row_indexer, col_indexer = indexer
            else:
                row_indexer = slice(None)
                col_indexer = indexer

            has_col_indexer = col_indexer is not None

        if col_indexer is None:
            col_indexer = dataset.column_names

        super().__init__()
        self._row_indexer = row_indexer
        self._col_indexer = col_indexer
        self._has_col_indexer = has_col_indexer

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._row_indexer}, {self._col_indexer}, {self._has_col_indexer})"

    @property
    def row(self) -> Union[int, slice, Iterable[int], Iterable[bool], IntegralTensor]:
        return self._row_indexer

    @property
    def col(self) -> Union[int, str, Iterable[int], Iterable[str]]:
        return self._col_indexer

    @property
    def single_row(self) -> bool:
        return isinstance(self._row_indexer, (int,))

    @property
    def single_col(self) -> bool:
        return isinstance(self._col_indexer, (int, str))

    @property
    def has_col_indexer(self) -> bool:
        return self._has_col_indexer


def _get_from_idx_indices_slice_mask(
    x: Union[Tensor, np.ndarray, pw.SupportsGetitemLen],
    row_indexer: Union[int, Tensor, np.ndarray, slice, Iterable[int], Iterable[bool]],
) -> Any:
    if isinstance(row_indexer, int):
        return x[row_indexer]
    elif isinstance(row_indexer, (SignedIntegerTensor0D, np.integer)):
        return x[row_indexer.item()]

    elif isinstance(row_indexer, slice):
        if isinstance(x, DynamicItemDataset):
            indices = range(len(x))[row_indexer]
            return [x[idx] for idx in indices]
        elif isinstance(x, Mapping):
            keys = list(x.keys())[row_indexer]
            return {k: x[k] for k in keys}
        elif isinstance(x, (Tensor, np.ndarray, tuple, list)):
            return x[row_indexer]
        else:
            msg = f"Invalid argument type {type(x)=} with {type(row_indexer)=}."
            raise TypeError(msg)

    elif pw.isinstance_generic(
        row_indexer, (Iterable[bool], BoolTensor1D)
    ) or is_numpy_bool_array(row_indexer):
        return _get_from_mask(x, row_indexer)  # type: ignore
    elif pw.isinstance_generic(
        row_indexer, (Iterable[int], SignedIntegerTensor1D)
    ) or is_numpy_integral_array(row_indexer):
        return _get_from_indices(x, row_indexer)  # type: ignore
    else:
        msg = f"Invalid argument type {type(row_indexer)=}."
        raise TypeError(msg)


def _get_from_indices(
    x: Union[Tensor, np.ndarray, pw.SupportsGetitemLen],
    indices: Union[Tensor, np.ndarray, Iterable],
) -> Union[Tensor, np.ndarray, list]:
    if isinstance(x, Tensor) and isinstance(indices, Tensor):
        return x[indices]
    elif isinstance(x, np.ndarray) and isinstance(indices, np.ndarray):
        return x[indices]
    elif isinstance(x, pw.SupportsGetitemLen):
        return [x[idx] for idx in indices]  # type: ignore
    else:
        msg = f"Invalid argument type {type(indices)=}."
        raise TypeError(msg)


def _get_from_mask(
    x: Union[Tensor, np.ndarray, pw.SupportsGetitemLen],
    mask: Union[Tensor, np.ndarray, Iterable],
) -> Union[Tensor, np.ndarray, list]:
    if isinstance(x, Tensor) and isinstance(mask, Tensor):
        return x[mask]
    elif isinstance(x, np.ndarray) and isinstance(mask, np.ndarray):
        return x[mask]
    elif isinstance(x, pw.SupportsGetitemLen):
        indices = multihot_to_multi_indices(mask)
        return [x[idx] for idx in indices]
    else:
        msg = f"Invalid argument type {type(mask)=}."
        raise TypeError(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Any, Dict, Literal, Optional, Tuple, Union, overload

from pythonwrench._core import _setup_output_fpath
from pythonwrench.inspect import get_fullname
from pythonwrench.typing.checks import isinstance_generic
from safetensors import safe_open
from safetensors.torch import save
from torch import Tensor

from torchwrench.nn import functional as F


@overload
def load_safetensors(
    fpath: Union[str, Path],
    *,
    device: str = "cpu",
    return_metadata: Literal[False] = False,
) -> Dict[str, Tensor]: ...


@overload
def load_safetensors(
    fpath: Union[str, Path],
    *,
    device: str = "cpu",
    return_metadata: Literal[True],
) -> Tuple[Dict[str, Tensor], Dict[str, str]]: ...


def load_safetensors(
    fpath: Union[str, Path],
    *,
    device: str = "cpu",
    return_metadata: bool = False,
) -> Union[Dict[str, Tensor], Tuple[Dict[str, Tensor], Dict[str, str]]]:
    tensors = {}
    with safe_open(fpath, framework="pt", device=device) as f:
        for k in f.keys():
            tensors[k] = f.get_tensor(k)

        if return_metadata:
            metadata = f.metadata()
            result = tensors, metadata
        else:
            result = tensors

    return result


@overload
def dump_safetensors(
    tensors: Dict[str, Tensor],
    fpath: Union[str, Path, None] = None,
    metadata: Optional[Dict[str, str]] = None,
    *,
    overwrite: bool = True,
    make_parents: bool = True,
    convert_to_tensor: Literal[False] = False,
) -> bytes: ...


@overload
def dump_safetensors(
    tensors: Dict[str, Any],
    fpath: Union[str, Path, None] = None,
    metadata: Optional[Dict[str, str]] = None,
    *,
    overwrite: bool = True,
    make_parents: bool = True,
    convert_to_tensor: Literal[True],
) -> bytes: ...


def dump_safetensors(
    tensors: Dict[str, Any],
    fpath: Union[str, Path, None] = None,
    metadata: Optional[Dict[str, str]] = None,
    *,
    overwrite: bool = True,
    make_parents: bool = True,
    convert_to_tensor: bool = False,
) -> bytes:
    """Dump tensors to safetensors format. Requires safetensors package installed."""
    if convert_to_tensor:
        tensors = {k: F.as_tensor(v) for k, v in tensors.items()}
    elif not isinstance_generic(tensors, Dict[str, Tensor]):
        msg = f"Invalid argument type {type(tensors)}. (expected dict[str, Tensor] but found {get_fullname(type(tensors))})"
        raise TypeError(msg)

    fpath = _setup_output_fpath(fpath, overwrite, make_parents)
    content = save(tensors, metadata)
    if fpath is not None:
        fpath.write_bytes(content)
    return content



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os
from pathlib import Path
from typing import (
    Any,
    BinaryIO,
    Callable,
    Dict,
    Optional,
    TextIO,
    TypeVar,
    Union,
    overload,
)

from pythonwrench.functools import function_alias
from pythonwrench.jsonl import load_jsonl
from typing_extensions import TypeAlias

from torchwrench.core.packaging import (
    _H5PY_AVAILABLE,
    _NUMPY_AVAILABLE,
    _SAFETENSORS_AVAILABLE,
    _TORCHAUDIO_AVAILABLE,
    _YAML_AVAILABLE,
)

from .common import SavingBackend, _fpath_to_saving_backend
from .csv import load_csv
from .json import load_json
from .pickle import load_pickle
from .torch import load_torch

T = TypeVar("T", covariant=True)
pylog = logging.getLogger(__name__)

LoadFn: TypeAlias = Callable[[Any], T]
LoadFnLike: TypeAlias = Union[LoadFn[T], SavingBackend]


LOAD_FNS: Dict[SavingBackend, LoadFn[Any]] = {
    "csv": load_csv,
    "json": load_json,
    "jsonl": load_jsonl,
    "pickle": load_pickle,
    "torch": load_torch,
}

if _H5PY_AVAILABLE:
    from .hdf import load_hdf

    LOAD_FNS["h5py"] = load_hdf


if _NUMPY_AVAILABLE:
    from .numpy import load_ndarray

    LOAD_FNS["numpy"] = load_ndarray


if _SAFETENSORS_AVAILABLE:
    from torchwrench.extras.safetensors import load_safetensors

    LOAD_FNS["safetensors"] = load_safetensors


if _TORCHAUDIO_AVAILABLE:
    from .torchaudio import load_with_torchaudio

    LOAD_FNS["torchaudio"] = load_with_torchaudio


if _YAML_AVAILABLE:
    from .yaml import load_yaml

    LOAD_FNS["yaml"] = load_yaml


@overload
def load(
    fpath: Union[TextIO, BinaryIO],
    *args,
    saving_backend: SavingBackend = "torch",
    **kwargs,
) -> Any: ...


@overload
def load(
    fpath: Union[str, Path, os.PathLike],
    *args,
    saving_backend: Optional[SavingBackend] = "torch",
    **kwargs,
) -> Any: ...


def load(
    fpath: Union[str, Path, os.PathLike, TextIO, BinaryIO],
    *args,
    saving_backend: Optional[SavingBackend] = "torch",
    **kwargs,
) -> Any:
    """Load from file using the correct backend."""
    if isinstance(fpath, (str, os.PathLike)):
        fpath = Path(fpath)

        if not fpath.is_file():
            msg = f"Invalid argument {fpath=}. (path is not a file)"
            raise FileNotFoundError(msg)

        if saving_backend is None:
            saving_backend = _fpath_to_saving_backend(fpath)

    if saving_backend not in LOAD_FNS:
        msg = f"Invalid argument {saving_backend=}. (expected one of {tuple(LOAD_FNS.keys())})"
        raise ValueError(msg)

    load_fn = LOAD_FNS[saving_backend]
    result = load_fn(fpath, *args, **kwargs)
    return result


@function_alias(load)
def read(*args, **kwargs): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torchwrench as tw
from torchwrench.utils.data.collate import AdvancedCollateDict


class TestCollate(TestCase):
    def test_advanced_example_1(self) -> None:
        collate = AdvancedCollateDict({"x": 0})

        inputs = [
            {"x": tw.as_tensor([1, 2])},
            {"x": tw.as_tensor([3, 4, 5, 6])},
            {"x": tw.as_tensor([7])},
        ]
        result = collate(inputs)
        expected = {"x": tw.as_tensor([[1, 2, 0, 0], [3, 4, 5, 6], [7, 0, 0, 0]])}
        assert tw.deep_equal(result, expected), f"{result=}"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Module versions of tensor functions that do not already exists in PyTorch."""

from typing import List, Optional, Sequence, Tuple, Union, overload

import torch
from pythonwrench.collections import dump_dict
from pythonwrench.semver import Version
from torch import Tensor
from torch.nn import functional as F
from torch.types import Number

from torchwrench.nn.functional.make import DTypeLike, as_dtype
from torchwrench.utils import return_types

from .module import Module


class Abs(Module):
    """
    Module version of :func:`~torch.abs`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.abs()


class Angle(Module):
    """
    Module version of :func:`~torch.angle`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.angle()


class Exp(Module):
    """
    Module version of :func:`~torch.exp`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.exp()


class Exp2(Module):
    """
    Module version of :func:`~torch.exp2`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.exp2()


class FFT(Module):
    """
    Module version of :func:`~torch.fft.fft`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return torch.fft.fft(x)


class IFFT(Module):
    """
    Module version of :func:`~torch.fft.ifft`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return torch.fft.ifft(x)


class Imag(Module):
    """
    Module version of :func:`~torch.Tensor.imag`.
    """

    def __init__(self, *, return_zeros: bool = False) -> None:
        """Return the imaginary part of a complex tensor.

        Args:
            return_zeros:
                If True and the input is not a complex tensor, the module will return a tensor of same shape containing zeros.
                If False and the input is not a complex tensor, raises the default RuntimError of PyTorch.
        """
        super().__init__()
        self.return_zeros = return_zeros

    def forward(self, x: Tensor) -> Tensor:
        if self.return_zeros and not x.is_complex():
            return torch.zeros_like(x)
        else:
            return x.imag


class Interpolate(Module):
    """
    Module version of :func:`~torch.nn.functional.interpolate`.
    """

    def __init__(
        self,
        size: Union[int, Tuple[int, ...], None] = None,
        scale_factor: Union[float, Tuple[float, ...], None] = None,
        mode: str = "nearest",
        align_corners: Optional[bool] = None,
        recompute_scale_factor: Optional[bool] = None,
        antialias: bool = False,
    ) -> None:
        super().__init__()
        self.size = size
        self.scale_factor = scale_factor
        self.mode = mode
        self.align_corners = align_corners
        self.recompute_scale_factor = recompute_scale_factor
        self.antialias = antialias

    def forward(self, x: Tensor) -> Tensor:
        kwds = {}
        if Version(torch.__version__) >= Version("2.0.0"):
            kwds.update(
                recompute_scale_factor=self.recompute_scale_factor,
                antialias=self.antialias,
            )

        return F.interpolate(
            x,
            size=self.size,
            scale_factor=self.scale_factor,
            mode=self.mode,
            align_corners=self.align_corners,
            **kwds,
        )


class Log(Module):
    """
    Module version of :func:`~torch.log`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.log()


class Log10(Module):
    """
    Module version of :func:`~torch.log10`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.log10()


class Log2(Module):
    """
    Module version of :func:`~torch.log2`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.log2()


class Max(Module):
    """
    Module version of :func:`~torch.max`.
    """

    def __init__(
        self,
        dim: Optional[int] = None,
        keepdim: bool = False,
        *,
        return_values: bool = True,
        return_indices: Optional[bool] = None,
    ) -> None:
        if return_indices is None:
            return_indices = dim is not None
        if not return_values and not return_indices:
            msg = f"Invalid combinaison of arguments {return_values=} and {return_indices=}. (at least one of them must be True)"
            raise ValueError(msg)
        if dim is None and keepdim:
            msg = f"Invalid combinaison of arguments {dim=} and {keepdim=}. (expected dim is not None or keepdim=False)"
            raise ValueError(msg)

        super().__init__()
        self.dim = dim
        self.return_values = return_values
        self.return_indices = return_indices
        self.keepdim = keepdim

    def forward(self, x: Tensor) -> Union[Tensor, return_types.max]:
        if self.dim is None:
            index = x.argmax()
            values_indices = return_types.max([x.flatten()[index], index])
        else:
            values_indices = x.max(dim=self.dim, keepdim=self.keepdim)

        if self.return_values and self.return_indices:
            return values_indices  # type: ignore
        elif self.return_values:
            return values_indices.values
        elif self.return_indices:
            return values_indices.indices
        else:
            msg = f"Invalid combinaison of arguments {self.return_values=} and {self.return_indices=}. (at least one of them must be True)"
            raise ValueError(msg)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
                return_values=self.return_values,
                return_indices=self.return_indices,
                keepdim=self.keepdim,
            ),
        )


class Mean(Module):
    """
    Module version of :func:`~torch.mean`.
    """

    def __init__(
        self,
        dim: Optional[int] = None,
        keepdim: bool = False,
        dtype: DTypeLike = None,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.keepdim = keepdim
        self.dtype = dtype

    def forward(self, x: Tensor) -> Tensor:
        dtype = as_dtype(self.dtype)
        if (Version(torch.__version__) >= Version("2.0.0")) or (self.dim is not None):
            return x.mean(dim=self.dim, keepdim=self.keepdim, dtype=dtype)  # type: ignore

        # support for older torch versions
        result = x.mean(dtype=dtype)
        if self.keepdim:
            return torch.full(x.shape, result.item(), dtype=dtype, device=x.device)
        else:
            return result

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
                keepdim=self.keepdim,
                dtype=self.dtype,
            ),
            ignore_lst=(None,),
        )


class Min(Module):
    """
    Module version of :func:`~torch.min`.
    """

    def __init__(
        self,
        dim: Optional[int] = None,
        keepdim: bool = False,
        *,
        return_values: bool = True,
        return_indices: Optional[bool] = None,
    ) -> None:
        if return_indices is None:
            return_indices = dim is not None
        if not return_values and not return_indices:
            msg = f"Invalid combinaison of arguments {return_values=} and {return_indices=}. (at least one of them must be True)"
            raise ValueError(msg)
        if dim is None and keepdim:
            msg = f"Invalid combinaison of arguments {dim=} and {keepdim=}. (expected dim is not None or keepdim=False)"
            raise ValueError(msg)

        super().__init__()
        self.dim = dim
        self.return_values = return_values
        self.return_indices = return_indices
        self.keepdim = keepdim

    def forward(self, x: Tensor) -> Union[Tensor, return_types.min]:
        if self.dim is None:
            index = x.argmin()
            values_indices = return_types.min([x.flatten()[index], index])
        else:
            values_indices = x.min(dim=self.dim, keepdim=self.keepdim)

        if self.return_values and self.return_indices:
            return values_indices  # type: ignore
        elif self.return_values:
            return values_indices.values
        elif self.return_indices:
            return values_indices.indices
        else:
            msg = f"Invalid combinaison of arguments {self.return_values=} and {self.return_indices=}. (at least one of them must be True)"
            raise ValueError(msg)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim=self.dim,
                return_values=self.return_values,
                return_indices=self.return_indices,
                keepdim=self.keepdim,
            ),
        )


class Normalize(Module):
    """
    Module version of :func:`~torch.nn.functional.normalize`.
    """

    def __init__(
        self,
        p: float = 2.0,
        dim: int = 1,
        eps: float = 1e-12,
    ) -> None:
        super().__init__()
        self.p = p
        self.dim = dim
        self.eps = eps

    def forward(self, x: Tensor) -> Tensor:
        return F.normalize(x, self.p, self.dim, self.eps)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                p=self.p,
                dim=self.dim,
                eps=self.eps,
            )
        )


class Permute(Module):
    """
    Module version of :func:`~torch.permute`.
    """

    def __init__(self, *args: int) -> None:
        super().__init__()
        self.dims = tuple(args)

    def forward(self, x: Tensor) -> Tensor:
        return x.permute(self.dims)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dims=self.dims,
            ),
            fmt="{value}",
        )


class Pow(Module):
    """
    Module version of :func:`~torch.Tensor.pow`.
    """

    def __init__(self, exponent: Union[Number, Tensor]) -> None:
        super().__init__()
        self.exponent = exponent

    def forward(self, x: Tensor) -> Tensor:
        return x.pow(self.exponent)

    def extra_repr(self) -> str:
        return dump_dict(exponent=self.exponent)


class Real(Module):
    """
    Module version of :func:`~torch.Tensor.real`.
    """

    def forward(self, x: Tensor) -> Tensor:
        return x.real


class Repeat(Module):
    """
    Module version of :func:`~torch.repeat`.
    """

    def __init__(self, *repeats: int) -> None:
        super().__init__()
        self.repeats = repeats

    def forward(self, x: Tensor) -> Tensor:
        return x.repeat(self.repeats)

    def extra_repr(self) -> str:
        return dump_dict(repeats=self.repeats)


class RepeatInterleave(Module):
    """
    Module version of :func:`~torch.repeat_interleave`.
    """

    def __init__(
        self,
        repeats: Union[int, Tensor],
        dim: int,
        output_size: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.repeats = repeats
        self.dim = dim
        self.output_size = output_size

    def forward(self, x: Tensor) -> Tensor:
        return x.repeat_interleave(self.repeats, self.dim, output_size=self.output_size)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                repeats=self.repeats,
                dim=self.dim,
                output_size=self.output_size,
            ),
            ignore_lst=(None,),
        )


class Reshape(Module):
    """
    Module version of :func:`~torch.reshape`.
    """

    def __init__(self, *shape: int) -> None:
        super().__init__()
        self.shape = shape

    def forward(self, x: Tensor) -> Tensor:
        return x.reshape(self.shape)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                shape=self.shape,
            ),
        )


class Sort(Module):
    """
    Module version of :func:`~torch.Tensor.sort`.
    """

    def __init__(
        self,
        dim: int = -1,
        descending: bool = False,
        *,
        return_values: bool = True,
        return_indices: bool = True,
    ) -> None:
        if not return_values and not return_indices:
            msg = f"Invalid combinaison of arguments {return_values=} and {return_indices=}. (at least one of them must be True)"
            raise ValueError(msg)

        super().__init__()
        self.dim = dim
        self.descending = descending
        self.return_values = return_values
        self.return_indices = return_indices

    def forward(self, x: Tensor) -> Union[return_types.sort, Tensor]:
        result = x.sort(dim=self.dim, descending=self.descending)
        result = return_types.sort(result)

        if self.return_values and self.return_indices:
            return result
        elif self.return_values:
            return result.values
        elif self.return_indices:
            return result.indices
        else:
            msg = f"Invalid combinaison of arguments {self.return_values=} and {self.return_indices=}. (at least one of them must be True)"
            raise ValueError(msg)


class TensorTo(Module):
    """
    Module version of :func:`~torch.Tensor.to`.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__()
        self.kwargs = kwargs

    def forward(self, x: Tensor) -> Tensor:
        return x.to(**self.kwargs)

    def extra_repr(self) -> str:
        return dump_dict(self.kwargs)


class ToList(Module):
    """
    Module version of :func:`~torch.Tensor.tolist`.
    """

    def forward(self, x: Tensor) -> List:
        return x.tolist()


class Transpose(Module):
    """
    Module version of :func:`~torch.transpose`.
    """

    def __init__(self, dim0: int, dim1: int, copy: bool = False) -> None:
        super().__init__()
        self.dim0 = dim0
        self.dim1 = dim1
        self.copy = copy

    def forward(self, x: Tensor) -> Tensor:
        if self.copy and not hasattr(torch, "transpose_copy"):
            msg = f"Invalid argument {self.copy=} in torch {torch.__version__}."
            raise ValueError(msg)

        if self.copy:
            return torch.transpose_copy(x, self.dim0, self.dim1)  # type: ignore
        else:
            return torch.transpose(x, self.dim0, self.dim1)

    def extra_repr(self) -> str:
        return dump_dict(
            dict(
                dim0=self.dim0,
                dim1=self.dim1,
            ),
            fmt="{value}",
        )


class View(Module):
    @overload
    def __init__(self, dtype: torch.dtype, /) -> None: ...

    @overload
    def __init__(self, size: Sequence[int], /) -> None: ...

    @overload
    def __init__(self, *size: int) -> None: ...

    def __init__(self, *args) -> None:
        super().__init__()
        self.args = args

    def forward(self, x: Tensor) -> Tensor:
        return x.view(*self.args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import TypeVar, Union

import torch
from pythonwrench.typing.classes import BuiltinNumber, BuiltinScalar
from typing_extensions import TypeAlias

from torchwrench.extras.numpy.definitions import (  # noqa: F401
    NumpyNumberLike,
    NumpyScalarLike,
    np,
)

from .tensor_subclasses import (  # noqa: F401
    BoolTensor,
    BoolTensor0D,
    BoolTensor1D,
    BoolTensor2D,
    BoolTensor3D,
    ByteTensor,
    ByteTensor0D,
    ByteTensor1D,
    ByteTensor2D,
    ByteTensor3D,
    CDoubleTensor,
    CDoubleTensor0D,
    CDoubleTensor1D,
    CDoubleTensor2D,
    CDoubleTensor3D,
    CFloatTensor,
    CFloatTensor0D,
    CFloatTensor1D,
    CFloatTensor2D,
    CFloatTensor3D,
    CharTensor,
    CharTensor0D,
    CharTensor1D,
    CharTensor2D,
    CharTensor3D,
    ComplexFloatingTensor,
    ComplexFloatingTensor0D,
    ComplexFloatingTensor1D,
    ComplexFloatingTensor2D,
    ComplexFloatingTensor3D,
    DoubleTensor,
    DoubleTensor0D,
    DoubleTensor1D,
    DoubleTensor2D,
    DoubleTensor3D,
    FloatingTensor,
    FloatingTensor0D,
    FloatingTensor1D,
    FloatingTensor2D,
    FloatingTensor3D,
    FloatTensor,
    FloatTensor0D,
    FloatTensor1D,
    FloatTensor2D,
    FloatTensor3D,
    HalfTensor,
    HalfTensor0D,
    HalfTensor1D,
    HalfTensor2D,
    HalfTensor3D,
    IntegralTensor,
    IntegralTensor0D,
    IntegralTensor1D,
    IntegralTensor2D,
    IntegralTensor3D,
    IntTensor,
    IntTensor0D,
    IntTensor1D,
    IntTensor2D,
    IntTensor3D,
    LongTensor,
    LongTensor0D,
    LongTensor1D,
    LongTensor2D,
    LongTensor3D,
    ShortTensor,
    ShortTensor0D,
    ShortTensor1D,
    ShortTensor2D,
    ShortTensor3D,
    SignedIntegerTensor,
    SignedIntegerTensor0D,
    SignedIntegerTensor1D,
    SignedIntegerTensor2D,
    SignedIntegerTensor3D,
    Tensor0D,
    Tensor1D,
    Tensor2D,
    Tensor3D,
    UnsignedIntegerTensor,
    UnsignedIntegerTensor0D,
    UnsignedIntegerTensor1D,
    UnsignedIntegerTensor2D,
    UnsignedIntegerTensor3D,
)

r"""/!\ The following type hints are meant for type annotation only, not for runtime checks.
"""

NumberLike: TypeAlias = Union[BuiltinNumber, NumpyNumberLike, Tensor0D]
ScalarLike: TypeAlias = Union[BuiltinScalar, NumpyScalarLike, Tensor0D]
TensorOrArray: TypeAlias = Union[torch.Tensor, np.ndarray]

T_Tensor = TypeVar("T_Tensor", bound=torch.Tensor, covariant=True)
T_TensorOrArray = TypeVar(
    "T_TensorOrArray",
    bound=TensorOrArray,
    covariant=True,
)

del np



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import itertools
from typing import Iterable, Iterator, List, Literal, Sequence, Union

import torch
from pythonwrench.semver import Version
from torch import Tensor
from torch.utils.data.sampler import Sampler

from torchwrench.nn.functional.transform import (
    GeneratorLike,
    as_generator,
    as_tensor,
    shuffled,
)


class SubsetSampler(Sampler[int]):
    def __init__(self, indices: Union[List[int], Tensor]) -> None:
        """
        A sampler to load a subset of a dataset from indices.

        Args:
            indices: List of indices to return.
        """
        if Version(str(torch.__version__)) < "2.10.0":
            args = (None,)
        else:
            args = ()

        indices = as_tensor(indices, dtype="long")

        super().__init__(*args)
        self._indices = indices

    def __iter__(self) -> Iterator[int]:
        return iter(self._indices.tolist())

    def __len__(self) -> int:
        return len(self._indices)


class SubsetCycleSampler(Sampler[int]):
    def __init__(
        self,
        indices: Union[Tensor, Iterable[int]],
        n_max_iterations: Union[int, Literal["inf"]] = "inf",
        shuffle: bool = True,
        seed: GeneratorLike = None,
    ) -> None:
        """SubsetCycleSampler that cycle on indices indifinitely or until a number max of iterations is reached.

        Args:
            indices: The list of indices of the items.
            n_max_iterations: The maximal number of iterations.
                If "inf", any call to __len__ will raises a NotImplementedError exception.
                defaults to "inf".
            shuffle: If True, shuffle the indices at every len(indices).
                defaults to True.
            seed: Optional seed or generator used to shuffle indices.
                defaults to None.
        """
        if Version(str(torch.__version__)) < "2.10.0":
            args = (None,)
        else:
            args = ()

        indices = as_tensor(indices, dtype="long")
        generator = as_generator(seed)

        super().__init__(*args)
        self._indices = indices
        self._n_max_iterations = n_max_iterations
        self._shuffle = shuffle
        self._generator = generator

        self._shuffle_indices()

    def __iter__(self) -> Iterator[int]:
        for i, idx in enumerate(itertools.cycle(self._indices)):
            if i % len(self._indices) == len(self._indices) - 1:
                self._shuffle_indices()

            if isinstance(self._n_max_iterations, int) and i >= self._n_max_iterations:
                break

            yield idx.item()  # type: ignore

    def __len__(self) -> int:
        if isinstance(self._n_max_iterations, int):
            return self._n_max_iterations
        elif self._n_max_iterations == "inf":
            msg = "Infinite sampler does not have __len__() method."
            raise NotImplementedError(msg)
        else:
            msg = f"Invalid argument {self._n_max_iterations=}."
            raise ValueError(msg)

    def _shuffle_indices(self) -> None:
        if not self._shuffle:
            return None
        self._indices = shuffled(self._indices, generator=self._generator)


class BalancedSampler(Sampler):
    def __init__(
        self,
        indices_per_class: Sequence[Sequence[int]],
        n_max_iterations: int,
        shuffle: bool = True,
        seed: GeneratorLike = None,
    ) -> None:
        """BalancedSampler class.

        Args:
            indices_per_class: List of indices per class index.
            n_max_iterations: The maximal number of iterations.
                If "inf", any call to __len__ will raises a NotImplementedError exception.
                defaults to "inf".
            shuffle: If True, shuffle the indices at every len(indices).
                defaults to True.
            seed: Optional seed or generator used to shuffle indices.
                defaults to None.
        """
        if Version(str(torch.__version__)) < "2.10.0":
            args = (None,)
        else:
            args = ()

        indices_per_class = [list(indices) for indices in indices_per_class]
        for cls_idx, indices in enumerate(indices_per_class):
            if len(indices) == 0:
                msg = f"Found a class index {cls_idx} without any indices."
                raise RuntimeError(msg)

        max_idx = max(len(indices) for indices in indices_per_class)
        pointers_per_class = [
            list(range(len(indices))) for indices in indices_per_class
        ]
        local_idx_per_class = [0 for _ in range(len(indices_per_class))]
        generator = as_generator(seed)

        super().__init__(*args)
        self._cls_to_sample_indices = indices_per_class
        self._n_max_iterations = n_max_iterations
        self._shuffle = shuffle
        self._generator = generator

        self._max_idx = max_idx
        self._pointers_per_class = pointers_per_class
        self._local_idx_per_class = local_idx_per_class

        self._shuffle_indices()

    def __iter__(self) -> Iterator[int]:
        global_idx = 0
        n_classes = len(self._cls_to_sample_indices)
        cls_indices = torch.randperm(n_classes, generator=self._generator)

        for i, cls_idx in enumerate(itertools.cycle(range(n_classes))):
            cls_idx = int(cls_indices[cls_idx].item())
            if global_idx >= self._n_max_iterations:
                break
            if global_idx % self._max_idx == self._max_idx - 1:
                self._shuffle_indices()

            sample_indices = self._cls_to_sample_indices[cls_idx]
            pointers = self._pointers_per_class[cls_idx]
            pointer_idx = self._local_idx_per_class[cls_idx]

            pointer = pointers[pointer_idx]
            sample_idx = sample_indices[pointer]

            yield sample_idx

            self._local_idx_per_class[cls_idx] = (pointer_idx + 1) % len(pointers)
            global_idx += 1

            if self._shuffle and i == n_classes - 1:
                cls_indices = shuffled(cls_indices, generator=self._generator)

    def __len__(self) -> int:
        return self._n_max_iterations

    def _shuffle_indices(self) -> None:
        if not self._shuffle:
            return None

        for i, pointers in enumerate(self._pointers_per_class):
            pointers_pt = as_tensor(pointers)
            pointers_pt = shuffled(pointers_pt, generator=self._generator)
            self._pointers_per_class[i] = pointers_pt.tolist()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time
import unittest
import warnings
from unittest import TestCase
from urllib.error import HTTPError

from torch import Tensor

from torchwrench.hub.download import _get_filename_from_url
from torchwrench.hub.paths import get_tmp_dir
from torchwrench.hub.registry import RegistryHub


class TestFilename(TestCase):
    def test_filename_from_url(self) -> None:
        tests = [
            (
                "https://dl.fbaipublicfiles.com/pyslowfast/dataset/class_names/kinetics_classnames.json",
                "kinetics_classnames.json",
            ),
            (
                "https://zenodo.org/record/8020843/files/convnext_tiny_465mAP_BL_AC_70kit.pth?download=1",
                "convnext_tiny_465mAP_BL_AC_70kit.pth",
            ),
            ("random.test", "random.test"),
        ]

        for input_, expected in tests:
            result = _get_filename_from_url(input_, timeout_s=10.0)
            assert result == expected


class TestRegistryHub(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        tmpdir = get_tmp_dir().joinpath("torchwrench_tests")
        tmpdir.mkdir(parents=True, exist_ok=True)
        cls.tmpdir = tmpdir

    def test_cnext_register(self) -> None:
        register = RegistryHub(
            infos={
                "cnext_bl_70": {
                    "architecture": "ConvNeXt",
                    "url": "https://zenodo.org/record/8020843/files/convnext_tiny_465mAP_BL_AC_70kit.pth?download=1",
                    "hash_value": "0688ae503f5893be0b6b71cb92f8b428",
                    "hash_type": "md5",
                    "fname": "convnext_tiny_465mAP_BL_AC_70kit.pth",
                    "state_dict_key": "model",
                },
            },
            register_root=self.tmpdir,
        )

        model_name = "cnext_bl_70"
        num_tries = 5
        for i in range(num_tries):
            try:
                register.download_file(model_name, force=False)
                break
            except HTTPError as err:
                if i + 1 < num_tries:
                    msg = f"Failed to download {i + 1} times."
                    warnings.warn(msg)
                    raise err
                else:
                    msg = f"Found error {err} at try number {i + 1}/{num_tries}, retrying..."
                    warnings.warn(msg)
                    time.sleep(i + 1)

        state_dict = register.load_state_dict(
            model_name,
            offline=True,
            load_kwds=dict(map_location="cpu", weights_only=False),
        )

        assert isinstance(state_dict, dict)
        assert all(isinstance(k, str) for k in state_dict.keys())
        assert all(isinstance(v, Tensor) for v in state_dict.values())


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch
from pythonwrench.functools import function_alias
from pythonwrench.os import get_num_cpus_available


@function_alias(torch.cuda.device_count)
def get_num_gpus_available(*args, **kwargs): ...


@function_alias(get_num_cpus_available)
def get_auto_num_cpus(*args, **kwargs): ...


@function_alias(get_num_gpus_available)
def get_auto_num_gpus(*args, **kwargs): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pythonwrench.jsonl import dump_jsonl, load_jsonl
from torch.serialization import *  # type: ignore

from .common import as_builtin
from .csv import dump_csv, load_csv
from .dump_fn import dump, save
from .json import dump_json, load_json
from .load_fn import load
from .pickle import dump_pickle, load_pickle
from .torch import dump_torch, load_torch



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .yaml import (
    IgnoreTagLoader,
    SplitTagLoader,
    YamlLoaders,
    dump_yaml,
    dumps_yaml,
    load_yaml,
    loads_yaml,
    read_yaml,
    save_yaml,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import warnings
from io import BytesIO
from pathlib import Path
from typing import Literal, Optional, Union, get_args

from pythonwrench._core import _setup_output_fpath
from pythonwrench.functools import function_alias
from pythonwrench.warnings import deprecated_alias

from .definitions import np

NumpyFormat = Literal["npy", "npz"]


def dump_ndarray(
    obj: np.ndarray,
    fpath: Union[str, Path, None] = None,
    *args,
    np_format: Optional[NumpyFormat] = "npy",
    overwrite: bool = True,
    make_parents: bool = True,
    **kwargs,
) -> bytes:
    fpath = _setup_output_fpath(fpath, overwrite, make_parents)

    if np_format is not None:
        pass
    elif fpath is None or fpath.suffix == ".npy":
        np_format = "npy"
    elif fpath.suffix == ".npz":
        np_format = "npz"
    else:
        msg = f"Unknown numpy extension '{fpath.suffix}'. (expected one of {get_args(NumpyFormat)})"
        warnings.warn(msg)
        np_format = "npy"

    if np_format == "npy":
        save_fn = np.save
    elif np_format == "npz":
        save_fn = np.savez
    else:
        raise ValueError(f"Invalid argument {np_format=}.")

    buffer = BytesIO()
    save_fn(buffer, obj, *args, **kwargs)
    buffer.seek(0)
    content = buffer.read()

    if fpath is not None:
        fpath.write_bytes(content)
    return content


@deprecated_alias(dump_ndarray)
def dump_numpy(*args, **kwargs): ...


@function_alias(dump_ndarray)
def dumps_ndarray(*args, **kwargs): ...


@function_alias(dump_ndarray)
def save_ndarray(*args, **kwargs): ...


def load_ndarray(fpath: Union[str, Path], *args, **kwargs) -> np.ndarray:
    return np.load(fpath, *args, **kwargs)


@function_alias(load_ndarray)
def read_ndarray(*args, **kwargs): ...


@function_alias(load_ndarray)
def loads_ndarray(*args, **kwargs): ...


@deprecated_alias(load_ndarray)
def load_numpy(*args, **kwargs): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import unittest
from collections import Counter
from pathlib import Path
from typing import Any, Dict, List, Tuple, get_args
from unittest import TestCase

import pythonwrench as pw
import torch

import torchwrench as tw
from torchwrench.core.packaging import (
    _NUMPY_AVAILABLE,
    _PANDAS_AVAILABLE,
    _SAFETENSORS_AVAILABLE,
    _YAML_AVAILABLE,
)
from torchwrench.hub.paths import get_tmp_dir
from torchwrench.nn.functional import deep_equal
from torchwrench.serialization.common import (
    SavingBackend,
    _fpath_to_saving_backend,
    as_builtin,
)


class TestSaving(TestCase):
    def test_examples(self) -> None:
        x = [
            [
                torch.arange(3)[None],
                "a",
                Path("./path"),
                Counter(["a", "b", "a", "c", "a"]),
                (),
            ],
        ]
        expected = [[[list(range(3))], "a", "path", {"a": 3, "b": 1, "c": 1}, []]]

        if _PANDAS_AVAILABLE:
            import pandas as pd

            df = pd.DataFrame({"a": [1, 2]})

            x += [
                df,
                df["a"],
            ]
            expected += [
                {"a": [1, 2]},
                [1, 2],
            ]

        result = as_builtin(x)
        assert result == expected, f"{result=}; {expected=}"

    def test_as_builtin(self) -> None:
        assert as_builtin(tw.arange(10)) == list(range(10))

        with self.assertRaises(TypeError):
            as_builtin(self)

    def test_detect_backend(self) -> None:
        tests: List[Tuple[str, str]] = [
            ("test.json", "json"),
            ("test.yaml.json", "json"),
        ]
        if _YAML_AVAILABLE:
            tests += [
                ("test.json.yaml", "yaml"),
                ("test.yml", "yaml"),
            ]

        for fpath, expected_backend in tests:
            backend = _fpath_to_saving_backend(fpath)
            assert backend == expected_backend

    def test_csv(self) -> None:
        data = {
            "a": [pw.randstr(10) for _ in range(100)],
            "b": [pw.randstr(10) for _ in range(100)],
        }

        fpath = get_tmp_dir().joinpath("tmp.csv")
        tw.dump_csv(data, fpath)
        result = tw.load_csv(fpath, orient="dict")

        assert result == data

        if _NUMPY_AVAILABLE and _PANDAS_AVAILABLE:
            import numpy as np

            n = 10
            data_matrix = dict(zip(pw.randstr(n), tw.rand(n, 100).numpy()))

            tw.dump_csv(data_matrix, fpath, backend="pandas")
            result = tw.load_csv(fpath, orient="dict", backend="pandas")

            assert all(np.allclose(v, result[k]) for k, v in data_matrix.items())

    def test_save_load(self) -> None:
        n = 1
        data_tensors = {
            "arange": tw.arange(n),
            "full": tw.full((n, 5), 9),
            "ones": tw.ones(n, 5),
            "rand": tw.rand(n),
            "randint": tw.randint(0, 100, (n,)),
            "randperm": tw.randperm(n),
            "zeros": tw.zeros(n, 1),
            "empty": tw.empty(n, 2),
        }
        data_objs: Dict[str, Any] = copy.copy(data_tensors)
        data_objs.update({"randstr": [pw.randstr(2) for _ in range(n)]})

        assert pw.is_full(map(len, data_objs.values()))

        tests: List[Tuple[str, Any, bool, dict, dict]] = [
            ("json", data_objs, True, dict(), dict()),
            ("pickle", data_objs, False, dict(), dict()),
        ]

        if _NUMPY_AVAILABLE:
            from torchwrench.extras.numpy import to_ndarray

            tests += [
                ("numpy", to_ndarray(v), False, dict(), dict())
                for k, v in data_objs.items()
            ]

        if _SAFETENSORS_AVAILABLE:
            tests += [
                ("safetensors", data_tensors, False, dict(), dict()),
            ]

        if _YAML_AVAILABLE:
            tests += [
                ("yaml", data_tensors, True, dict(), dict()),
                ("yaml", data_objs, True, dict(), dict()),
            ]

        for i, (backend, data, to_builtins, load_kwds, dump_kwds) in enumerate(tests):
            assert tw.isinstance_generic(backend, SavingBackend), (  # type: ignore
                f"{backend=}; {get_args(SavingBackend)=}"
            )

            if to_builtins:
                data = tw.as_builtin(data)
            if backend == "safetensors":
                # note: safetensors automatically sort dicts
                data = pw.sorted_dict(data)

            fpath = get_tmp_dir().joinpath(f"tmp.{backend}")
            tw.dump(data, fpath, saving_backend=backend, **dump_kwds)
            result = tw.load(fpath, saving_backend=backend, **load_kwds)

            assert deep_equal(data, result), f"{backend=}, {i=}/{len(tests)}"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from pathlib import Path
from typing import Any, BinaryIO, Callable, Dict, Optional, Union, overload

from pythonwrench.functools import function_alias
from pythonwrench.jsonl import dump_jsonl
from typing_extensions import TypeAlias

from torchwrench.core.packaging import (
    _H5PY_AVAILABLE,
    _NUMPY_AVAILABLE,
    _SAFETENSORS_AVAILABLE,
    _TORCHAUDIO_AVAILABLE,
    _YAML_AVAILABLE,
)

from .common import SavingBackend, _fpath_to_saving_backend
from .csv import dump_csv
from .json import dump_json
from .pickle import dump_pickle
from .torch import dump_torch

DumpFn: TypeAlias = Callable[..., Any]
DumpFnLike: TypeAlias = Union[DumpFn, SavingBackend]


DUMP_FNS: Dict[SavingBackend, DumpFn] = {
    "csv": dump_csv,
    "json": dump_json,
    "jsonl": dump_jsonl,
    "pickle": dump_pickle,
    "torch": dump_torch,
}

if _H5PY_AVAILABLE:
    from .hdf import dump_hdf

    DUMP_FNS["h5py"] = dump_hdf

if _NUMPY_AVAILABLE:
    from .numpy import dump_ndarray

    DUMP_FNS["numpy"] = dump_ndarray

if _SAFETENSORS_AVAILABLE:
    from torchwrench.extras.safetensors import dump_safetensors

    DUMP_FNS["safetensors"] = dump_safetensors

if _TORCHAUDIO_AVAILABLE:
    from .torchaudio import dump_with_torchaudio

    DUMP_FNS["torchaudio"] = dump_with_torchaudio

if _YAML_AVAILABLE:
    from .yaml import dump_yaml

    DUMP_FNS["yaml"] = dump_yaml


@overload
def dump(
    obj: Any,
    fpath: Union[None, BinaryIO] = None,
    *args,
    saving_backend: SavingBackend = "torch",
    **kwargs,
) -> Union[str, bytes]: ...


@overload
def dump(
    obj: Any,
    fpath: Union[str, Path, os.PathLike],
    *args,
    saving_backend: Optional[SavingBackend] = "torch",
    **kwargs,
) -> Union[str, bytes]: ...


def dump(
    obj: Any,
    fpath: Union[str, Path, os.PathLike, None, BinaryIO] = None,
    *args,
    saving_backend: Optional[SavingBackend] = "torch",
    **kwargs,
) -> Union[str, bytes]:
    """Load from file using the correct backend."""
    if isinstance(fpath, (str, os.PathLike)):
        fpath = Path(fpath)

    if saving_backend is None:
        if not isinstance(fpath, (str, Path, os.PathLike)):
            msg = f"Invalid combinaison of arguments {fpath=} and {saving_backend=}."
            raise ValueError(msg)

        saving_backend = _fpath_to_saving_backend(fpath)

    elif saving_backend not in DUMP_FNS:
        msg = f"Invalid argument {saving_backend=}. (expected one of {tuple(DUMP_FNS.keys())})"
        raise ValueError(msg)

    dump_fn = DUMP_FNS[saving_backend]
    result = dump_fn(obj, fpath, *args, **kwargs)
    return result


@function_alias(dump)
def save(*args, **kwargs): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
from typing import Callable, Dict, Generic, Iterable, List, Mapping, Optional, overload

from torch import nn
from typing_extensions import Concatenate, ParamSpec

from torchwrench.core.config import _REPLACE_MODULE_CLASSES

from ._mixins import (
    _DEFAULT_DEVICE_DETECT_MODE,
    ConfigModule,
    DeviceDetectMode,
    EModule,  # noqa: F401
    ESequential,  # noqa: F401
    InType,
    OutType,
    OutType3,
    TypedModuleLike,
)

P = ParamSpec("P")


class EModuleList(
    Generic[InType, OutType3],
    EModule[InType, List[OutType3]],
    nn.ModuleList,
):
    """Enriched torch.nn.ModuleList with proxy device, forward typing and automatic configuration detection from attributes.

    Designed to work with `torchwrench.nn.EModule` instances.
    The default behaviour is the same than PyTorch ModuleList class, except for the forward call which returns a list containing the output of each module called separately.
    """

    @overload
    def __init__(
        self,
        modules: Optional[Iterable[TypedModuleLike[InType, OutType3]]] = None,
        *,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None: ...

    @overload
    def __init__(
        self,
        modules: Optional[Iterable[nn.Module]] = None,
        *,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None: ...

    def __init__(
        self,
        modules=None,
        *,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        EModule.__init__(
            self,
            strict_load=strict_load,
            config_to_extra_repr=config_to_extra_repr,
            device_detect_mode=device_detect_mode,
        )
        nn.ModuleList.__init__(self, modules)

    def forward(self, *args: InType, **kwargs: InType) -> List[OutType3]:
        return [module(*args, **kwargs) for module in self]


class EModuleDict(
    Generic[InType, OutType3],
    EModule[InType, Dict[str, OutType3]],
    nn.ModuleDict,
):
    """Enriched torch.nn.ModuleDict with proxy device, forward typing and automatic configuration detection from attributes.

    Designed to work with `torchwrench.nn.EModule` instances.
    The default behaviour is the same than PyTorch ModuleDict class, except for the forward call which returns a dict containing the output of each module called separately.
    """

    @overload
    def __init__(
        self,
        modules: Optional[Mapping[str, TypedModuleLike[InType, OutType3]]] = None,
        *,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None: ...

    @overload
    def __init__(
        self,
        modules: Optional[Mapping[str, nn.Module]] = None,
        *,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None: ...

    def __init__(
        self,
        modules=None,
        *,
        strict_load: bool = False,
        config_to_extra_repr: bool = False,
        device_detect_mode: DeviceDetectMode = _DEFAULT_DEVICE_DETECT_MODE,
    ) -> None:
        EModule.__init__(
            self,
            strict_load=strict_load,
            config_to_extra_repr=config_to_extra_repr,
            device_detect_mode=device_detect_mode,
        )
        nn.ModuleDict.__init__(self, modules)

    def forward(self, *args: InType, **kwargs: InType) -> Dict[str, OutType3]:
        return {name: module(*args, **kwargs) for name, module in self.items()}  # type: ignore


class EModulePartial(
    Generic[InType, OutType],
    EModule[InType, OutType],
):
    """Wrap a python callable to nn.Module class."""

    def __init__(
        self,
        fn: Callable[Concatenate[InType, P], OutType],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        """
        Args:
            fn: Callable to wrap.
            *args: Positional arguments.
            **kwargs:
        """
        super().__init__()
        self.fn = fn
        self.args = args
        self.kwargs = kwargs

    def forward(self, x: InType, **kwargs: P.kwargs) -> OutType:  # type: ignore
        kwds = copy.copy(self.kwargs)
        kwds.update(kwargs)
        return self.fn(x, *self.args, **kwds)  # type: ignore

    def extra_repr(self) -> str:
        return f"{self.fn.__name__}, {ConfigModule.extra_repr(self)}"


if _REPLACE_MODULE_CLASSES:
    ModuleList = EModuleList
    ModuleDict = EModuleDict
    Sequential = ESequential
else:
    from torch.nn import ModuleDict, ModuleList, Sequential  # noqa: F401

ModulePartial = EModulePartial



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

import torch

from torchwrench.core.dtype_enum import DTypeEnum


class TestDTypeEnum(TestCase):
    def test_dtype_enum_inst(self) -> None:
        dt1 = DTypeEnum.float16

        assert dt1.dtype == torch.float16 == torch.half
        assert dt1 == "float16"
        assert dt1 == "half"
        assert dt1 == "torch.float16"
        assert dt1 == "torch.half"

    def test_check_properties(self) -> None:
        dt = torch.complex128
        dte = DTypeEnum.from_dtype(dt)

        assert dt.is_complex == dte.is_complex
        assert dt.is_floating_point == dte.is_floating_point
        assert dt.is_signed == dte.is_signed

        try:
            expected = dt.itemsize  # type: ignore
        except AttributeError as err:
            expected = err

        try:
            result = dte.itemsize
        except AttributeError as err:
            result = err

        assert (
            isinstance(expected, AttributeError) and isinstance(result, AttributeError)
        ) or (expected == result)

        try:
            expected = dt.to_real()  # type: ignore
        except AttributeError as err:
            expected = err

        try:
            result = dte.to_real()
        except AttributeError as err:
            result = err

        assert (
            isinstance(expected, AttributeError) and isinstance(result, AttributeError)
        ) or (expected == result)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pythonwrench.pickle import (  # noqa: F401
    dump_pickle,
    dumps_pickle,
    load_pickle,
    loads_pickle,
    read_pickle,
    save_pickle,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pythonwrench.importlib import Placeholder
from pythonwrench.inspect import get_current_fn_name


class BaseLoader(Placeholder): ...


class CBaseLoader(Placeholder): ...


class CFullLoader(Placeholder): ...


class CLoader(Placeholder): ...


class CSafeLoader(Placeholder): ...


class CUnsafeLoader(Placeholder): ...


class FullLoader(Placeholder): ...


class Loader(Placeholder): ...


class MappingNode(Placeholder): ...


class Node(Placeholder): ...


class SafeLoader(Placeholder): ...


class ScalarNode(Placeholder): ...


class SequenceNode(Placeholder): ...


class UnsafeLoader(Placeholder): ...


class ParserError(RuntimeError): ...


class ScannerError(RuntimeError): ...


def load(*args, **kwargs):
    msg = f"Cannot call function '{get_current_fn_name()}' because optional dependency 'pyyaml' is not installed. Please install it using 'pip install torchwrench[extras]'"
    raise NotImplementedError(msg)


def dump(*args, **kwargs):
    msg = f"Cannot call function '{get_current_fn_name()}' because optional dependency 'pyyaml' is not installed. Please install it using 'pip install torchwrench[extras]'"
    raise NotImplementedError(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any

from pythonwrench.importlib import Placeholder


class DynamicItemDataset(Placeholder):
    def __len__(self) -> int:
        raise NotImplementedError

    def __getitem__(self, *args, **kwargs) -> Any:
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import inspect
import io
import os
import pickle
from io import BufferedWriter
from pathlib import Path
from typing import IO, Any, BinaryIO, Callable, Dict, Optional, Union

import torch
from pythonwrench._core import _setup_output_fpath
from pythonwrench.semver import Version
from torch.serialization import DEFAULT_PROTOCOL
from torch.types import Storage
from typing_extensions import TypeAlias

FileLike: TypeAlias = Union[str, os.PathLike, BinaryIO, IO[bytes]]
MapLocationLike: TypeAlias = Optional[
    Union[Callable[[Storage, str], Storage], torch.device, str, Dict[str, str]]
]


def dump_torch(
    obj: object,
    f: Optional[FileLike] = None,
    pickle_module: Any = pickle,
    pickle_protocol: int = DEFAULT_PROTOCOL,
    _use_new_zipfile_serialization: bool = True,
    _disable_byteorder_record: bool = False,
    *,
    overwrite: bool = True,
    make_parents: bool = True,
) -> bytes:
    if "_disable_byteorder_record" in inspect.getargs(torch.save.__code__).args:
        kwds = dict(_disable_byteorder_record=_disable_byteorder_record)
    else:
        kwds = {}

    buffer = io.BytesIO()
    torch.save(
        obj,
        buffer,
        pickle_module,
        pickle_protocol,
        _use_new_zipfile_serialization,
        **kwds,
    )
    content = buffer.getvalue()
    buffer.close()

    if isinstance(f, (str, Path, os.PathLike)) or f is None:
        f = _setup_output_fpath(f, overwrite, make_parents)

    if isinstance(f, Path):
        f.write_bytes(content)
    elif isinstance(f, (BinaryIO, BufferedWriter)):
        f.write(content)
        f.flush()

    return content


def load_torch(
    f: FileLike,
    map_location: MapLocationLike = None,
    pickle_module: Any = None,
    *,
    weights_only: bool = ...,
    mmap: Optional[bool] = None,
    **pickle_load_args: Any,
) -> Any:
    kwds = {}

    if Version(torch.__version__) < Version("2.1.0"):
        pickle_module = pickle
    else:
        if weights_only is ...:
            weights_only = Version(torch.__version__) >= "2.6.0"

        kwds.update(
            weights_only=weights_only,
            mmap=mmap,
        )

    return torch.load(
        f,
        map_location,
        pickle_module,
        **kwds,
        **pickle_load_args,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torch.hub import *  # type: ignore

from .download import download_file
from .paths import get_cache_dir, get_tmp_dir
from .registry import RegistryEntry, RegistryHub



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    NamedTuple,
    Sized,
    Tuple,
    Union,
    get_args,
)

import torch
from torch import Size, Tensor
from torch.types import Number
from typing_extensions import TypeAlias

from torchwrench.core.make import (
    DeviceLike,
    DTypeLike,
    GeneratorLike,
    as_device,
    as_dtype,
    as_generator,
)
from torchwrench.nn import functional as F
from torchwrench.types import Tensor0D, is_number_like

PadAlign: TypeAlias = Literal["left", "right", "center", "random"]
PadValue: TypeAlias = Union[Number, Tensor0D, Callable[[Tensor], Number]]
PadMode: TypeAlias = Literal["constant", "reflect", "replicate", "circular"]


class PaddedValue(NamedTuple):
    padded: Tensor
    lengths: Tensor


def pad_dim(
    x: Tensor,
    target_length: int,
    *,
    dim: int = -1,
    align: PadAlign = "left",
    pad_value: PadValue = 0.0,
    mode: PadMode = "constant",
    generator: GeneratorLike = None,
) -> Tensor:
    """Generic function to pad a tensor along a single dimension.

    Args:
        x: Tensor to pad of with N dims of shape (..., D, ...), where D is the size of the dim-th dimension.
        target_length: Target length for dim.
        dims: Axis/dim to pad. defaults to -1.
        align: Alignement for pad.
        pad_value: Pad value used to fill tensor. defaults to 0.0.
            Note: Complex values is currently not supported.
        mode: Pad mode used. defaults to "constant".
        generator: Random generator when align is "random".

    Returns:
        Padded tensor of N dims of (..., target_length, ...).
    """
    return pad_dims(
        x,
        target_lengths=[target_length],
        dims=[dim],
        aligns=[align],
        pad_value=pad_value,
        mode=mode,
        generator=generator,
    )


def pad_dims(
    x: Tensor,
    target_lengths: Iterable[int],
    *,
    dims: Union[Iterable[int], Literal["auto"], None] = None,
    aligns: Union[PadAlign, Iterable[PadAlign]] = "left",
    pad_value: PadValue = 0.0,
    mode: PadMode = "constant",
    generator: GeneratorLike = None,
) -> Tensor:
    """Generic function to pad a tensor along multiple dimensions.

    Args:
        x: Tensor to pad of with N dims.
        target_lengths: List of target lengths for each dimension. The list has size M <= N.
        dims: Dimensions for each length. Must be of size M. If "auto", creates a list of the M last dimensions.
        aligns: Alignement or list of alignements for each dimension of size M.
        pad_value: Pad value used to fill tensor. defaults to 0.0.
            Note: Complex values is currently not supported.
        mode: Pad mode used. defaults to "constant".
        generator: Random generator when aligns is "random".

    Returns:
        Padded tensor of N dims.
    """
    if isinstance(pad_value, Callable):
        pad_value = pad_value(x)

    target_lengths = list(target_lengths)

    if isinstance(aligns, str):
        aligns_lst = [aligns] * len(target_lengths)  # type: ignore
    else:
        aligns_lst = list(aligns)
    del aligns

    if dims == "auto" or dims is None:
        dims = list(range(-len(target_lengths), 0))
    else:
        dims = list(dims)

    if len(target_lengths) != len(dims):
        msg = f"Invalid number of targets lengths ({len(target_lengths)}) with the number of dimensions ({len(dims)})."
        raise ValueError(msg)

    if len(aligns_lst) != len(dims):
        msg = f"Invalid number of aligns ({len(aligns_lst)}) with the number of dimensions ({len(dims)})."
        raise ValueError(msg)

    pad_seq = __generate_pad_seq(
        x.shape,
        target_lengths=target_lengths,
        dims=dims,
        aligns=aligns_lst,  # type: ignore
        generator=generator,
    )
    x = F.pad(x, pad_seq, mode=mode, value=pad_value)  # type: ignore
    return x


def pad_and_stack_rec(
    sequence: Union[Tensor, int, float, tuple, list],
    pad_value: Number = 0,
    *,
    align: PadAlign = "left",
    device: DeviceLike = None,
    dtype: DTypeLike = None,
) -> Tensor:
    """Recursive version of torch.nn.utils.rnn.pad_sequence, with padding of Tensors.

    Args:
        sequence: The sequence to pad. Must be convertable to tensor by having the correct number of dims in all sublists.
        pad_value: The pad value used. defaults to 0.
            Note: Complex values is currently not supported.
        align: Alignement used for padding. defaults to "left".
        device: The device of the output Tensor. defaults to None.
        dtype: The dtype of the output Tensor. defaults to None.

    Example 1::
    -----------
    >>> sequence = [[1, 2], [3], [], [4, 5]]
    >>> output = pad_sequence_rec(sequence, 0)
    tensor([[1, 2], [3, 0], [0, 0], [4, 5]])

    Example 2::
    -----------
    >>> invalid_sequence = [[1, 2, 3], 3]
    >>> output = pad_sequence_rec(invalid_sequence, 0)
    ValueError : Cannot pad sequence of tensors of differents number of dims.

    """
    device = as_device(device)
    dtype = as_dtype(dtype)

    def _impl(sequence: Union[Tensor, int, float, tuple, list]) -> Tensor:
        if isinstance(sequence, Tensor):
            return sequence.to(dtype=dtype, device=device)

        elif is_number_like(sequence) or (
            isinstance(sequence, Sized) and len(sequence) == 0
        ):
            return torch.as_tensor(sequence, dtype=dtype, device=device)  # type: ignore

        elif isinstance(sequence, (list, tuple)):
            tensors = [_impl(elt) for elt in sequence]
            if F.is_stackable(tensors):
                return torch.stack(tensors)

            shapes = [elt.shape for elt in tensors]
            shape0 = shapes[0]

            if not all(len(shape) == len(shape0) for shape in shapes[1:]):
                msg = f"Cannot pad sequence of tensors of differents number of dims. (with {shapes=})"
                raise ValueError(msg)

            max_lens = [max(shape[i] for shape in shapes) for i in range(len(shape0))]
            tensors = [
                pad_dims(
                    xi,
                    target_lengths=max_lens,
                    pad_value=pad_value,
                    aligns=[align] * xi.ndim,  # type: ignore
                    dims=range(xi.ndim),
                )
                for xi in tensors
            ]
            result = torch.stack(tensors)
            return result

        else:
            msg = f"Invalid type {type(sequence)}. (expected Tensor, int, float, list or tuple)"
            raise TypeError(msg)

    return _impl(sequence)


def collate_tensors(
    tensors: Iterable[Tensor],
    pad_value: float = 0.0,
    dim: int = -1,
) -> PaddedValue:
    tensors = list(tensors)
    if len(tensors) == 0:
        return PaddedValue(torch.empty((0,)), torch.empty((0,)))

    device = tensors[0].device
    lengths = torch.as_tensor([tensor.shape[dim] for tensor in tensors], device=device)
    maxlength = int(lengths.max().item())
    tensors = [
        pad_dim(tensor, maxlength, dim=dim, pad_value=pad_value) for tensor in tensors
    ]
    padded = torch.stack(tensors)
    return PaddedValue(padded, lengths)


def cat_padded_batch(
    x1: Tensor,
    x1_lens: Tensor,
    x2: Tensor,
    x2_lens: Tensor,
    seq_dim: int = -1,
    batch_dim: int = 0,
) -> Tuple[Tensor, Tensor]:
    """Concatenate padded batched of sequences.

    Args:
        x1: First batch with D dims of shape (batch_size, ..., N1, ...)
        x1_lens: First lengths of each element in sequence dim of shape (batch_size,).
        x2: Second batch with D dims of shape (batch_size, ..., N2, ...)
            The shape must be the same than x1 unless for the dimension N2.
        x2_lens: Second lengths of each element in sequence dim of shape (batch_size,).
        seq_dim: Dimension index of sequence. defaults to -1.
        batch_dim: Batch dimension index. defaults to 0.
    """
    _check_cat_padded_batch(x1, x1_lens, x2, x2_lens, seq_dim, batch_dim)
    x12_lens = x1_lens + x2_lens
    sum_size_12 = x1.shape[seq_dim] + x2.shape[seq_dim]

    x12 = pad_dim(x1, sum_size_12, dim=seq_dim)
    kwd: Dict[str, Any] = dict(device=x1.device, dtype=torch.long)
    indices = torch.arange(x2_lens.max().item(), **kwd)

    unsq_x1_lens = x1_lens
    ndim = x1.ndim
    for i in range(ndim):
        if i != (seq_dim % ndim):
            indices = indices.unsqueeze(dim=i)
        if i != (batch_dim % ndim):
            unsq_x1_lens = unsq_x1_lens.unsqueeze(dim=i)

    expand_size = list(x2.shape)
    expand_size[seq_dim] = -1
    indices = indices.expand(*expand_size)
    indices = indices + unsq_x1_lens
    x12.scatter_(seq_dim, indices, x2)

    max_size_12 = int(x12_lens.max().item())
    if max_size_12 < sum_size_12:
        slices = [slice(None) for _ in range(ndim)]
        slices[seq_dim] = slice(max_size_12)
        x12 = x12[tuple(slices)]

    return x12, x12_lens


def __generate_pad_seq(
    x_shape: Union[Size, Tuple[int, ...]],
    target_lengths: Iterable[int],
    *,
    dims: Iterable[int] = (-1,),
    aligns: Iterable[PadAlign] = ("left",),
    generator: GeneratorLike = None,
) -> List[int]:
    """Generate pad sequence for torch.nn.functional.pad from target lengths, dims, aligns and generator.

    Args:
        x_shape: Shape of the tensor to pad.
        target_lengths: Expected lengths of each dim in the tensor.
        dims: Specified dims indices. defaults to (-1,).
        aligns: Alignment for each dim. defaults to ("left",).
        generator: Optional generator when align="random". defaults to None.
    """
    target_lengths = list(target_lengths)
    aligns = list(aligns)
    dims = list(dims)
    generator = as_generator(generator)

    if len(dims) == 0:
        msg = f"Invalid argument {dims=}. (cannot use an empty list of dimensions)"
        raise ValueError(msg)

    if len(target_lengths) != len(dims):
        msg = f"Invalid number of targets lengths ({len(target_lengths)}) with the number of dimensions ({len(dims)})."
        raise ValueError(msg)

    if len(aligns) != len(dims):
        msg = f"Invalid number of aligns ({len(aligns)}) with the number of dimensions ({len(dims)})."
        raise ValueError(msg)

    pad_seq = [0 for _ in range(len(x_shape) * 2)]
    for target_length, dim, align in zip(target_lengths, dims, aligns):
        missing = max(target_length - x_shape[dim], 0)

        if align == "left":
            missing_left = 0
            missing_right = missing
        elif align == "right":
            missing_left = missing
            missing_right = 0
        elif align == "center":
            missing_left = missing // 2 + missing % 2
            missing_right = missing // 2
        elif align == "random":
            missing_left = int(
                torch.randint(
                    low=0,
                    high=missing + 1,
                    size=(),
                    generator=generator,
                ).item()
            )
            missing_right = missing - missing_left
        else:
            msg = f"Invalid argument {align=}. (expected one of {get_args(PadAlign)})"
            raise ValueError(msg)

        # Note: pad_seq : [pad_left_dim_-1, pad_right_dim_-1, pad_left_dim_-2, pad_right_dim_-2, ...)
        idx = len(x_shape) - (dim % len(x_shape)) - 1
        assert pad_seq[idx * 2] == 0 and pad_seq[idx * 2 + 1] == 0
        pad_seq[idx * 2] = missing_left
        pad_seq[idx * 2 + 1] = missing_right

    return pad_seq


def _check_cat_padded_batch(
    x1: Tensor,
    x1_lens: Tensor,
    x2: Tensor,
    x2_lens: Tensor,
    seq_dim: int,
    batch_dim: int,
) -> None:
    if x1.ndim != x2.ndim:
        raise ValueError(f"Invalid arguments ndims. (found {x1.ndim=} != {x2.ndim=})")
    if x1.ndim < 2:
        raise ValueError(f"Invalid arguments ndims. (found {x1.ndim=} < 2)")

    batch_size = x1.shape[batch_dim]
    if not (x1_lens.shape == x2_lens.shape == Size((batch_size,))):
        msg = f"Invalid arguments shape. (with {x1_lens.shape=} and {x2_lens.shape=})"
        raise ValueError(msg)

    x1_shape = torch.as_tensor(x1.shape)
    x2_shape = torch.as_tensor(x2.shape)
    eq_mask = x1_shape.eq(x2_shape)
    eq_mask[seq_dim] = True
    if not eq_mask.all():
        raise ValueError(f"Invalid arguments shape. (with {x1.shape=} and {x2.shape=})")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Collection of functions and modules to help development in PyTorch."""

__name__ = "torchwrench"
__author__ = "Étienne Labbé (Labbeti)"
__author_email__ = "labbeti.pub@gmail.com"
__license__ = "MIT"
__maintainer__ = "Étienne Labbé (Labbeti)"
__status__ = "Development"
__version__ = "0.7.7"

# Import global functions and classes from torch
# Imports from pythonwrench
from pythonwrench.semver import Version
from pythonwrench.typing.checks import (
    is_builtin_collection,
    is_builtin_number,
    is_builtin_obj,
    is_builtin_scalar,
    is_dataclass_instance,
    is_namedtuple_instance,
    is_typed_dict,
    isinstance_generic,
)
from torch import *  # type: ignore
from torch import (
    complex64,
    complex128,
    double,
    float,
    float16,
    float32,
    float64,
    half,
    int,
    int8,
    int16,
    int32,
    int64,
    long,
    short,
    uint8,
)

# Re-import torchwrench modules for language servers
from . import core as core
from . import extras as extras
from . import hub as hub
from . import nn as nn
from . import optim as optim
from . import serialization as serialization
from . import special as special
from . import types as types
from . import utils as utils
from .core.dtype_enum import DTypeEnum
from .core.dtype_enum import DTypeEnum as dtype_enum
from .core.make import (
    CUDA_IF_AVAILABLE,
    Device,
    DeviceLike,
    DTypeLike,
    Generator,
    GeneratorLike,
)

# Global imports
from .core.version import get_githash_full, get_githash_short
from .extras.numpy import ndarray_to_tensor, tensor_to_ndarray, to_ndarray
from .extras.yaml.yaml import (
    dump_yaml,
    dumps_yaml,
    load_yaml,
    loads_yaml,
    read_yaml,
    save_yaml,
)
from .hub.download import download_file
from .hub.paths import get_cache_dir, get_tmp_dir
from .nn.functional import *
from .serialization.common import as_builtin
from .serialization.csv import (
    dump_csv,
    dumps_csv,
    load_csv,
    loads_csv,
    read_csv,
    save_csv,
)
from .serialization.dump_fn import dump, save
from .serialization.json import (
    dump_json,
    dumps_json,
    load_json,
    loads_json,
    read_json,
    save_json,
)
from .serialization.load_fn import load
from .serialization.pickle import (
    dump_pickle,
    dumps_pickle,
    load_pickle,
    loads_pickle,
    read_pickle,
    save_pickle,
)
from .serialization.torch import dump_torch, load_torch
from .types.guards import (
    is_number_like,
    is_scalar_like,
    is_tensor_or_array,
)
from .types.tensor_subclasses import (
    BoolTensor,
    BoolTensor0D,
    BoolTensor1D,
    BoolTensor2D,
    BoolTensor3D,
    ByteTensor,
    ByteTensor0D,
    ByteTensor1D,
    ByteTensor2D,
    ByteTensor3D,
    CDoubleTensor,
    CDoubleTensor0D,
    CDoubleTensor1D,
    CDoubleTensor2D,
    CDoubleTensor3D,
    CFloatTensor,
    CFloatTensor0D,
    CFloatTensor1D,
    CFloatTensor2D,
    CFloatTensor3D,
    CharTensor,
    CharTensor0D,
    CharTensor1D,
    CharTensor2D,
    CharTensor3D,
    ComplexFloatingTensor,
    ComplexFloatingTensor0D,
    ComplexFloatingTensor1D,
    ComplexFloatingTensor2D,
    ComplexFloatingTensor3D,
    DoubleTensor,
    DoubleTensor0D,
    DoubleTensor1D,
    DoubleTensor2D,
    DoubleTensor3D,
    FloatingTensor,
    FloatingTensor0D,
    FloatingTensor1D,
    FloatingTensor2D,
    FloatingTensor3D,
    FloatTensor,
    FloatTensor0D,
    FloatTensor1D,
    FloatTensor2D,
    FloatTensor3D,
    HalfTensor,
    HalfTensor0D,
    HalfTensor1D,
    HalfTensor2D,
    HalfTensor3D,
    IntegralTensor,
    IntegralTensor0D,
    IntegralTensor1D,
    IntegralTensor2D,
    IntegralTensor3D,
    IntTensor,
    IntTensor0D,
    IntTensor1D,
    IntTensor2D,
    IntTensor3D,
    LongTensor,
    LongTensor0D,
    LongTensor1D,
    LongTensor2D,
    LongTensor3D,
    RealTensor,
    RealTensor0D,
    RealTensor1D,
    RealTensor2D,
    RealTensor3D,
    ShortTensor,
    ShortTensor0D,
    ShortTensor1D,
    ShortTensor2D,
    ShortTensor3D,
    SignedIntegerTensor,
    SignedIntegerTensor0D,
    SignedIntegerTensor1D,
    SignedIntegerTensor2D,
    SignedIntegerTensor3D,
    Tensor,
    Tensor0D,
    Tensor1D,
    Tensor2D,
    Tensor3D,
    UnsignedIntegerTensor,
    UnsignedIntegerTensor0D,
    UnsignedIntegerTensor1D,
    UnsignedIntegerTensor2D,
    UnsignedIntegerTensor3D,
)

version = __version__
version_info = Version(__version__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any

from pythonwrench.importlib import Placeholder
from pythonwrench.inspect import get_current_fn_name


class DataFrame(Placeholder):
    def __getitem__(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def __setitem__(self, *args, **kwargs) -> Any:
        raise NotImplementedError


class Series(Placeholder): ...


class RangeIndex(Placeholder): ...


class Index(Placeholder): ...


def read_csv(*args, **kwargs) -> DataFrame:
    msg = f"Cannot call function '{get_current_fn_name()}' because optional dependency 'pandas' is not installed. Please install it using 'pip install torchwrench[extras]'"
    raise NotImplementedError(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Modules (classes) interface."""

from torch.nn.modules import *  # type: ignore

from .activation import LogSoftmaxMultidim, SoftmaxMultidim
from .container import (
    EModuleDict,
    EModuleList,
    EModulePartial,
    ESequential,
    ModuleDict,
    ModuleList,
    ModulePartial,
    Sequential,
)
from .crop import CropDim, CropDims
from .layer import PositionalEncoding
from .mask import MaskedMean, MaskedSum
from .module import EModule, Module
from .multiclass import (
    IndexToName,
    IndexToOnehot,
    NameToIndex,
    NameToOnehot,
    OnehotToIndex,
    OnehotToName,
    ProbsToIndex,
    ProbsToName,
    ProbsToOnehot,
)
from .multilabel import (
    IndicesToMultihot,
    IndicesToMultinames,
    MultihotToIndices,
    MultihotToMultiIndices,
    MultihotToMultinames,
    MultiIndicesToMultihot,
    MultiIndicesToMultinames,
    MultinamesToIndices,
    MultinamesToMultihot,
    MultinamesToMultiIndices,
    ProbsToIndices,
    ProbsToMultihot,
    ProbsToMultiIndices,
    ProbsToMultinames,
)
from .numpy import NDArrayToTensor, TensorToNDArray, ToNDArray
from .padding import PadAndStackRec, PadDim, PadDims
from .powerset import MultilabelToPowerset, PowersetToMultilabel
from .tensor import (
    FFT,
    IFFT,
    Abs,
    Angle,
    Exp,
    Exp2,
    Imag,
    Log,
    Log2,
    Log10,
    Max,
    Mean,
    Min,
    Normalize,
    Permute,
    Pow,
    Real,
    Repeat,
    RepeatInterleave,
    Reshape,
    Sort,
    TensorTo,
    ToList,
    Transpose,
    View,
)
from .transform import (
    AsTensor,
    Flatten,
    Identity,
    MoveToRec,
    PadAndCropDim,
    RepeatInterleaveNd,
    ResampleNearestFreqs,
    ResampleNearestRates,
    ResampleNearestSteps,
    Shuffled,
    Squeeze,
    ToItem,
    Topk,
    TopP,
    ToTensor,
    TransformDrop,
    Unsqueeze,
    ViewAsComplex,
    ViewAsReal,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from collections import Counter
from unittest import TestCase

from torchwrench.utils.data.sampler import BalancedSampler


class TestSampler(TestCase):
    def test_balanced_sampler(self) -> None:
        indices_per_class = [[0, 1], [2], [3, 4, 5]]

        for num_sampled_per_cls in (0, 10):
            num_classes = len(indices_per_class)

            n_max_iterations = num_classes * num_sampled_per_cls
            sampler = BalancedSampler(indices_per_class, n_max_iterations)

            def find_in_indices(x: int) -> int:
                for i, indices in enumerate(indices_per_class):
                    if x in indices:
                        return i
                return -1

            result = list(sampler)
            result_cls_indices = [find_in_indices(idx) for idx in result]

            result_counter = Counter(result_cls_indices)
            expected_counter = Counter(
                {i: n_max_iterations // num_classes for i in range(num_classes)}
                if num_sampled_per_cls > 0
                else {}
            )

            for i in range(0, len(result_cls_indices), num_classes):
                assert set(result_cls_indices[i : i + num_classes]) == set(
                    range(num_classes)
                )
            assert result_counter == expected_counter


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Module used for typing when numpy is not installed."""

from typing import Any

from pythonwrench.importlib import Placeholder
from pythonwrench.inspect import get_current_fn_name


class generic(Placeholder): ...


class number(generic): ...


class bool_(generic): ...


class str_(generic): ...


class dtype(Placeholder): ...


class ndarray(Placeholder):
    def __getitem__(self, *args) -> Any:
        return self


def array(x: Any, *args, **kwargs):
    msg = f"Cannot call function '{get_current_fn_name()}' because optional dependency 'numpy' is not installed. Please install it using 'pip install torchwrench[extras]'"
    raise NotImplementedError(msg)


def asarray(x: Any, *args, **kwargs):
    msg = f"Cannot call function '{get_current_fn_name()}' because optional dependency 'numpy' is not installed. Please install it using 'pip install torchwrench[extras]'"
    raise NotImplementedError(msg)


def iscomplexobj(x: Any):
    msg = f"Cannot call function '{get_current_fn_name()}' because optional dependency 'numpy' is not installed. Please install it using 'pip install torchwrench[extras]'"
    raise NotImplementedError(msg)


def empty(*args, **kwargs):
    msg = f"Cannot call function '{get_current_fn_name()}' because optional dependency 'numpy' is not installed. Please install it using 'pip install torchwrench[extras]'"
    raise NotImplementedError(msg)


class complex64(dtype): ...


class complex128(dtype): ...


class complex256(dtype): ...


class float16(dtype): ...


class float32(dtype): ...


class float64(dtype): ...


class float128(dtype): ...


class floating(dtype): ...


class int16(dtype): ...


class int32(dtype): ...


class int64(dtype): ...


class uint8(dtype): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import math
import re
from pathlib import Path
from typing import (
    Any,
    Dict,
    Literal,
    Optional,
    TypeVar,
    Union,
)

import torch
from pythonwrench.cast import as_builtin, register_as_builtin_fn
from torch import Tensor

from torchwrench.core.packaging import (
    _H5PY_AVAILABLE,
    _NUMPY_AVAILABLE,
    _OMEGACONF_AVAILABLE,
    _PANDAS_AVAILABLE,
    _SAFETENSORS_AVAILABLE,
    _TORCHAUDIO_AVAILABLE,
    _YAML_AVAILABLE,
)

T = TypeVar("T")

pylog = logging.getLogger(__name__)

SavingBackend = Literal[
    "csv",
    "json",
    "jsonl",
    "h5py",
    "numpy",
    "pickle",
    "safetensors",
    "torch",
    "torchaudio",
    "yaml",
]

# Note: order matter here: last extension of a backend is the default/recommanded one
PATTERN_TO_BACKEND: Dict[str, SavingBackend] = {
    r"^.+\.tsv$": "csv",
    r"^.+\.csv$": "csv",
    r"^.+\.json$": "json",
    r"^.+\.jsonl$": "jsonl",
    r"^.+\.pkl$": "pickle",
    r"^.+\.pickle$": "pickle",
    r"^.+\.torch$": "torch",
    r"^.+\.ckpt$": "torch",
    r"^.+\.pt$": "torch",
}

if _H5PY_AVAILABLE:
    PATTERN_TO_BACKEND.update(
        {
            r"^.+\.h5$": "h5py",
            r"^.+\.hdf$": "h5py",
            r"^.+\.hdf5$": "h5py",
        }
    )


if _NUMPY_AVAILABLE:
    import numpy as np

    PATTERN_TO_BACKEND.update(
        {
            r"^.+\.npz$": "numpy",
            r"^.+\.npy$": "numpy",
        }
    )

if _SAFETENSORS_AVAILABLE:
    PATTERN_TO_BACKEND.update(
        {
            r"^.+\.safetensors$": "safetensors",
        }
    )

if _TORCHAUDIO_AVAILABLE:
    PATTERN_TO_BACKEND.update(
        {
            r"^.+\.mp3$": "torchaudio",
            r"^.+\.wav$": "torchaudio",
            r"^.+\.aac$": "torchaudio",
            r"^.+\.ogg$": "torchaudio",
            r"^.+\.flac$": "torchaudio",
        }
    )

if _YAML_AVAILABLE:
    PATTERN_TO_BACKEND.update(
        {
            r".+\.yml$": "yaml",
            r".+\.yaml$": "yaml",
        }
    )


def _fpath_to_saving_backend(
    fpath: Union[str, Path],
    verbose: int = 0,
) -> SavingBackend:
    fname = Path(fpath).name

    saving_backend: Optional[SavingBackend] = None
    for pattern, backend in PATTERN_TO_BACKEND.items():
        if re.match(pattern, fname) is not None:
            saving_backend = backend
            break

    if saving_backend is None:
        msg = f"Unknown file pattern '{fname}'. (expected one of {tuple(PATTERN_TO_BACKEND.keys())} or specify the backend argument with `to.load(..., saving_backend=\"backend\")`)"
        raise ValueError(msg)

    if verbose >= 2:
        msg = f"Loading file '{str(fpath)}' using {saving_backend=}."
        pylog.debug(msg)
    return saving_backend


BACKEND_TO_PATTERN: Dict[SavingBackend, str] = {
    backend: ext for ext, backend in PATTERN_TO_BACKEND.items()
}


@register_as_builtin_fn(Tensor)
def _tensor_to_builtin(x: Tensor) -> Any:
    return x.tolist()


@register_as_builtin_fn(torch.dtype)
def _torch_dtype_to_builtin(x: torch.dtype) -> Any:
    return str(x)


if _NUMPY_AVAILABLE:
    import numpy as np

    @register_as_builtin_fn(np.ndarray)
    def _np_ndarray_to_builtin(x: np.ndarray) -> Any:
        return x.tolist()

    @register_as_builtin_fn(np.generic)
    def _np_generic_to_builtin(x: np.generic) -> Any:
        return x.item()

    @register_as_builtin_fn(np.dtype)
    def _np_dtype_to_builtin(x: np.dtype) -> Any:
        return str(x)


if _OMEGACONF_AVAILABLE:
    from omegaconf import DictConfig, ListConfig, OmegaConf  # type: ignore

    @register_as_builtin_fn((DictConfig, ListConfig))
    def _omegaconf_to_builtin(x: Union[DictConfig, ListConfig]) -> Any:
        return as_builtin(OmegaConf.to_container(x, resolve=False, enum_to_str=True))  # type: ignore


if _PANDAS_AVAILABLE:
    import pandas as pd
    from pandas._libs.missing import NAType

    @register_as_builtin_fn(pd.DataFrame)
    def _dataframe_to_builtin(x: pd.DataFrame) -> Any:
        return as_builtin(x.to_dict("list"))

    @register_as_builtin_fn(pd.Series)
    def _series_to_builtin(x: pd.Series) -> Any:
        return as_builtin(x.to_list())

    @register_as_builtin_fn(NAType)
    def _na_to_builtin(x: NAType) -> float:
        return math.nan
